import datetime as dt

import pandas as pd
import hvplot.pandas
import panel as pn
import panel_material_ui as pmui

df = pd.read_csv('https://datasets.holoviz.org/penguins/v1/penguins.csv')

WIDTH = 260

pmui.widgets.base.MaterialWidget.width = WIDTH

menu_items = [
    {"label": "Home", "icon": "home"},
    {"label": "Catalog", "icon": "category"},
    {"label": "Checkout", "icon": "shopping_cart"},
]

color = pmui.RadioButtonGroup(options=pmui.COLORS[:-3], value='primary')
color_select = pmui.Select(options=pmui.COLORS[:-3], value='primary', color=color)
color.jslink(color, value='color')
color.jslink(color_select, value='value', bidirectional=True)

switch = pmui.BreakpointSwitcher(
    breakpoint='md',
    small=color_select,
    large=color
)

drawer = pmui.Drawer('## Drawer')

pmui.Paper(
    pmui.Row(
        pmui.Typography('Components', variant='h4'),
        switch,
        pmui.ThemeToggle(styles={'margin-left': 'auto'}),
        sizing_mode='stretch_width'
    ),
    pmui.Tabs(
        ('Widgets', pn.FlexBox(
            pmui.Column(
                pmui.Typography('Sliders', variant='h6'),
                pmui.FloatSlider(label='FloatSlider', value=42.314, color=color),
                pmui.DateSlider(label='DateSlider', color=color, start=dt.datetime(2025, 4, 1), end=dt.datetime(2025, 4, 30), value=dt.datetime(2025, 4, 14)),
                pmui.EditableFloatSlider(label='EditableSlider', color=color, value=42.31, width=WIDTH),
                pmui.RangeSlider(label='RangeSlider', color=color, value=(13, 42))
            ),
            pmui.Column(
                pmui.Typography('Select', variant='h6'),
                pmui.Select(label='Select', options=pmui.COLORS, color=color, value='primary', width=WIDTH),
                pmui.MultiChoice(label='MultiChoice', options=pmui.COLORS, color=color, value=['primary']),
                pmui.MultiSelect(label='MultiSelect', color=color, options=pmui.COLORS, value=['primary'])
            ),
            pmui.Column(
                pmui.Typography('Text', variant='h6'),
                pmui.TextInput(label='TextInput', color=color, value='My Text'),
                pmui.PasswordInput(label='PasswordInput', color=color, value='mysupersecretpassword'),
                pmui.TextAreaInput(label='TextArea', color=color, value='1. Default\n2. Primary\n3. Secondary', rows=4)
            ),
            pmui.Column(
                pmui.Typography('Date & Time', variant='h6'),
                pmui.DatePicker(label='DatePicker', color=color, width=WIDTH),
                pmui.DatetimePicker(label='DatetimePicker', color=color, width=WIDTH),
                pmui.TimePicker(label='TimePicker', color=color)
            ),
            pmui.Column(
                pmui.Typography('Buttons', variant='h6'),
                pmui.Row(
                    pmui.Button(label='Button', color=color, icon='home'),
                    pmui.Button(label='Button', color=color, icon='home', variant='outlined')
                ),
                pmui.Row(
                    pmui.Fab(label='Fab', color=color, icon='home', variant='circular'),
                    pmui.Fab(label='Fab', color=color, icon='home', variant='extended')
                ),
                pmui.Row(
                    pmui.FileInput(label='FileInput', color=color),
                    pmui.FileInput(label='FileInput', color=color, variant='outlined')
                )
            ),
            pmui.Column(
                pmui.Typography('Menus', variant='h6'),
                pmui.Breadcrumbs(active=1, color=color, items=menu_items),
                pmui.List(active=1, color=color, items=menu_items),
                pmui.SpeedDial(active=1, color=color, items=menu_items),
                margin=10
            )
        )),
        ('Plots', pmui.Column(
            df.hvplot.scatter(
                x="bill_length_mm", y="bill_depth_mm", by="species",
                height=400, responsive=True, max_width=800
            ),
        )),
        ('Layouts', pn.FlexBox(
            pmui.Accordion(
                ('Accordion 1', 'Content'),
                ('Accordion 2', 'Content'),
                width=WIDTH, margin=(0, 10), active=[2]
            ),
            pmui.Card(
                'Card Content',
                title='Card',
                width=WIDTH, margin=(0, 10)
            ),
            pmui.Card(
                'Card Content',
                title='Card',
                width=WIDTH, margin=(0, 10)
            ),
            pmui.Column(
                pmui.Typography('Drawer', variant='h6'),
                pmui.RadioButtonGroup.from_param(drawer.param.anchor),
                pmui.Toggle.from_param(drawer.param.open),
                drawer,
                width=WIDTH, margin=(0, 10)
            ),
            pmui.Column(
                pmui.Typography('Paper', variant='h6'),
                pmui.Row(
                    pmui.Paper(width=50, height=50, margin=10, elevation=1),
                    pmui.Paper(width=50, height=50, margin=10, elevation=3),
                    pmui.Paper(width=50, height=50, margin=10, elevation=10),
                )
            ),
            margin=10
        )),
        color=color, sizing_mode='stretch_both',
    ),
    margin=5,
    sizing_mode="stretch_both",
    styles={'overflow': 'hidden'}
).save(
    'doc/_static/demo.html',
    resources='cdn'
)



import panel as pn

from panel_material_ui.widgets import Button, Toggle


def test_button(document, comm):
    button = Button(label='Test Button')
    assert button.label == 'Test Button'
    widget = button.get_root(document, comm=comm)
    assert isinstance(widget, pn.models.esm.ReactComponent)
    button._process_events({'clicks': 1})
    assert button.clicks == 1


def test_button_event():
    button = Button(name='Button')

    events = []
    def callback(event):
        events.append(event.new)

    button.param.watch(callback, 'value')
    assert button.value == False

    event = {"clicks": 1, "value": True}
    button._process_events(event)
    assert events == [True]
    assert button.value == False


def test_button_jscallback_clicks(document, comm):
    button = Button(name='Button')
    code = 'console.log("Clicked!")'
    button.jscallback(clicks=code)

    widget = button.get_root(document, comm=comm)
    assert len(widget.js_event_callbacks) == 1
    callbacks = widget.js_event_callbacks
    assert 'dom_event' in callbacks
    assert len(callbacks['dom_event']) == 1
    assert code in callbacks['dom_event'][0].code


def test_toggle(document, comm):
    toggle = Toggle(label='Test Toggle', value=True)
    assert toggle.value == True
    assert toggle.label == 'Test Toggle'

    widget = toggle.get_root(document, comm=comm)
    assert isinstance(widget, pn.models.esm.ReactComponent)

    toggle._process_events({'value': False})
    assert toggle.value == False
    toggle._process_events({'value': True})
    assert toggle.value == True



import pytest

pytest.importorskip('playwright')

from panel.layout import Column
from panel.widgets import Button
from panel.tests.util import serve_component, wait_until
from panel_material_ui.layout import Accordion
from playwright.sync_api import expect

pytestmark = pytest.mark.ui

def test_accordion_active(page):
    layout = Accordion(("Card 1", "Card 1 objects"), ("Card 2", "Card 2 objects"))
    serve_component(page, layout)

    expect(page.locator('.accordion')).to_have_count(1)
    # 2 collapsed cards
    cards = page.locator('.MuiAccordion-gutters')
    expect(cards).to_have_count(2)
    expanded_cards = page.locator('.MuiCollapse-entered')
    expect(expanded_cards).to_have_count(0)

    # expand the card, `active` is set properly
    card0 = cards.nth(0)
    card0.click()
    expect(expanded_cards).to_have_count(1)
    expanded_cards.wait_for(timeout=5000)
    assert layout.active == [0]
    # set `active`, the cards expanded accordingly
    layout.active = [0, 1]
    expect(page.locator('.MuiAccordionSummary-root.Mui-expanded')).to_have_count(2)

def test_accordion_basic(page):
    content1 = Column("Content 1")
    content2 = Column("Content 2")
    widget = Accordion(
        ('Section 1', content1),
        ('Section 2', content2)
    )
    serve_component(page, widget)

    # Check sections exist
    sections = page.locator('.MuiAccordion-root')
    expect(sections).to_have_count(2)

    # Check headers
    headers = page.locator('.MuiAccordionSummary-content')
    expect(headers.nth(0)).to_contain_text('Section 1')
    expect(headers.nth(1)).to_contain_text('Section 2')

def test_accordion_expansion(page):
    content1 = Column("Content 1")
    content2 = Column("Content 2")
    widget = Accordion(
        ('Section 1', content1),
        ('Section 2', content2)
    )
    serve_component(page, widget)

    # Initially no sections should be expanded
    hidden = page.locator('.MuiCollapse-hidden')
    expect(hidden).to_have_count(2)

    # Click to expand first section
    page.locator('.MuiAccordionSummary-root').first.click()
    expect(hidden).to_have_count(1)
    wait_until(lambda: widget.active == [0], page)

def test_accordion_toggle_mode(page):
    content1 = Column("Content 1")
    content2 = Column("Content 2")
    widget = Accordion(
        ('Section 1', content1),
        ('Section 2', content2),
        toggle=True
    )
    serve_component(page, widget)

    # Expand first section
    summaries = page.locator('.MuiAccordionSummary-root')
    summaries.nth(0).click()
    wait_until(lambda: widget.active == [0], page)

    # Expand second section (should collapse first in toggle mode)
    summaries.nth(1).click()
    wait_until(lambda: widget.active == [1], page)
    hidden = page.locator('.MuiCollapse-hidden')
    expect(hidden).to_have_count(1)

def test_accordion_multiple_sections(page):
    content1 = Column("Content 1")
    content2 = Column("Content 2")
    widget = Accordion(
        ('Section 1', content1),
        ('Section 2', content2),
        toggle=False  # Allow multiple sections to be expanded
    )
    serve_component(page, widget)

    # Expand both sections
    summaries = page.locator('.MuiAccordionSummary-root')
    summaries.nth(0).click()
    summaries.nth(1).click()

    hidden = page.locator('.MuiCollapse-hidden')
    expect(hidden).to_have_count(0)
    wait_until(lambda: sorted(widget.active) == [0, 1], page)

def test_accordion_disabled_sections(page):
    content1 = Column("Content 1")
    content2 = Column("Content 2")
    widget = Accordion(
        ('Section 1', content1),
        ('Section 2', content2),
        disabled=[1]  # Disable second section
    )
    serve_component(page, widget)

    # Check disabled state
    disabled = page.locator('.MuiAccordionSummary-root.Mui-disabled')
    expect(disabled).to_have_count(1)

    # Try to click disabled section (should not expand)
    page.locator('.MuiAccordionSummary-root').nth(1).click(force=True)
    hidden = page.locator('.MuiCollapse-hidden')
    expect(hidden).to_have_count(2)

def test_accordion_initial_state(page):
    content1 = Column("Content 1")
    content2 = Column("Content 2")
    widget = Accordion(
        ('Section 1', content1),
        ('Section 2', content2),
        active=[0]  # First section initially expanded
    )
    serve_component(page, widget)

    # Check initial expansion
    hidden = page.locator('.MuiCollapse-hidden')
    expect(hidden).to_have_count(1)
    expect(page.locator('.MuiAccordionDetails-root').nth(0)).to_contain_text('Content 1')

def test_accordion_collapse(page):
    content1 = Column("Content 1")
    widget = Accordion(
        ('Section 1', content1),
        active=[0]  # Initially expanded
    )
    serve_component(page, widget)

    # Click to collapse
    page.locator('.MuiAccordionSummary-root').click()
    hidden = page.locator('.MuiCollapse-hidden')
    expect(hidden).to_have_count(1)
    wait_until(lambda: widget.active == [], page)

def test_accordion_square_variant(page):
    content1 = Column("Content 1")
    widget = Accordion(
        ('Section 1', content1),
        square=True
    )
    serve_component(page, widget)

    # Check if square class is applied
    accordion = page.locator('.MuiAccordion-root')
    expect(accordion).to_have_css('border-radius', '0px')



from __future__ import annotations

import param
from panel.util import edit_readonly, isIn
from panel.widgets.base import Widget
from panel.widgets.select import NestedSelect as _PnNestedSelect
from panel.widgets.select import Select as _PnSelect
from panel.widgets.select import SingleSelectBase as _PnSingleSelectBase
from panel.widgets.select import _MultiSelectBase as _PnMultiSelectBase

from ..base import COLORS
from .base import MaterialWidget
from .button import _ButtonLike


class MaterialSingleSelectBase(MaterialWidget, _PnSingleSelectBase):
    """
    Base class for Material UI single-select widgets.

    This class combines Material UI styling with Panel's single select functionality.
    It provides the foundation for widgets that allow selecting a single value from
    a list of options, such as dropdown selects and autocomplete inputs.

    This is an abstract base class and should not be used directly.
    """

    value = param.Parameter(default=None, allow_None=True)

    _allows_values = False

    __abstract = True


class MaterialMultiSelectBase(MaterialWidget, _PnMultiSelectBase):
    """
    Base class for Material UI multi-select widgets.

    This class combines Material UI styling with Panel's multi-select functionality.
    It provides the foundation for widgets that allow selecting multiple values from
    a list of options, such as multi-select dropdowns and autocomplete inputs.

    This is an abstract base class and should not be used directly.
    """

    value = param.List(default=[], allow_None=True)

    _rename = {"name": "name"}

    _allows_values = False

    __abstract = True


class AutocompleteInput(MaterialSingleSelectBase):
    """
    The `AutocompleteInput` widget allows searching and selecting a single value
    from a list of `options`.

    It falls into the broad category of single-value, option-selection widgets
    that provide a compatible API and include the  `Select`,
    `RadioBoxGroup` and `RadioButtonGroup` widgets.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/AutocompleteInput.html
    - https://panel.holoviz.org/reference/widgets/AutocompleteInput.html
    - https://mui.com/material-ui/react-autocomplete/

    :Example:

    >>> AutocompleteInput(
    ...     label='Study', options=['Biology', 'Chemistry', 'Physics'],
    ... )
    """

    case_sensitive = param.Boolean(default=True, doc="""
        Enable or disable case sensitivity.""")

    color = param.Selector(objects=COLORS, default="default")

    min_characters = param.Integer(default=2, doc="""
        The number of characters a user must type before
        completions are presented.""")

    placeholder = param.String(
        default="",
        doc="""
        Placeholder for empty input field.""",
    )

    restrict = param.Boolean(default=True, doc="""
        Set to False in order to allow users to enter text that is not
        present in the list of completion strings.""")

    search_strategy = param.Selector(default='starts_with',
        objects=['starts_with', 'includes'], doc="""
        Define how to search the list of completion strings. The default option
        `"starts_with"` means that the user's text must match the start of a
        completion string. Using `"includes"` means that the user's text can
        match any substring of a completion string.""")

    value_input = param.Parameter(
        default="",
        readonly=True,
        doc="""
        Initial or entered text value updated on every key press.""",
    )

    variant = param.Selector(objects=["filled", "outlined", "standard"], default="outlined")

    _allows_none = True

    _esm_base = "Autocomplete.jsx"

    _rename = {"name": "name"}

    def _process_property_change(self, msg):
        if 'value' in msg and msg['value'] is None:
            return msg
        if not self.restrict and 'value' in msg:
            try:
                return super()._process_property_change(msg)
            except Exception:
                return Widget._process_property_change(self, msg)
        return super()._process_property_change(msg)

    def _process_param_change(self, msg):
        if 'value' in msg and not self.restrict and not isIn(msg['value'], self.values):
            with param.parameterized.discard_events(self):
                props = super()._process_param_change(msg)
                self.value = props['value'] = msg['value']
        else:
            props = super()._process_param_change(msg)
        return props

    @param.depends('value', watch=True, on_init=True)
    def _sync_value_input(self):
        with edit_readonly(self):
            self.value_input = self.value


class _SelectDropdownBase(MaterialWidget):
    """
    Base class for Material UI dropdown based select widgets.

    This is an abstract base class and should not be used directly.
    """

    bookmarks = param.List(default=[], doc="List of bookmarked options")

    disabled_options = param.List(default=[], nested_refs=True, doc="""
        Optional list of ``options`` that are disabled, i.e. unusable and
        un-clickable. If ``options`` is a dictionary the list items have to
        correspond to the values in the options dictionary..""")

    filter_str = param.String(default="", doc="Filter string for the dropdown")

    filter_on_search = param.Boolean(default=True, doc="""
        Whether options are filtered or merely highlighted on search.""")

    dropdown_height = param.Integer(default=500, doc="Height of the dropdown menu")

    dropdown_open = param.Boolean(default=False, doc="Whether the dropdown is open")

    searchable = param.Boolean(default=False, doc="Whether the dropdown is searchable")

    value_label = param.String(doc="Custom label to describe the current option(s).")

    __abstract = True


class Select(MaterialSingleSelectBase, _PnSelect, _SelectDropdownBase):
    """
    The `Select` widget allows selecting a value from a list.

    It falls into the broad category of single-value, option-selection widgets
    that provide a compatible API and include the  `AutocompleteInput`,
    `RadioBoxGroup` and `RadioButtonGroup` widgets.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/Select.html
    - https://panel.holoviz.org/reference/widgets/Select.html
    - https://mui.com/material-ui/react-select/

    :Example:

    >>> Select(label='Study', options=['Biology', 'Chemistry', 'Physics'])
    """

    color = param.Selector(objects=COLORS, default="default")

    groups = param.Dict(default=None, nested_refs=True, doc="""
        Dictionary whose keys are used to visually group the options
        and whose values are either a list or a dictionary of options
        to select from. Mutually exclusive with ``options``  and valid only
        if ``size`` is 1.""")

    size = param.Integer(default=1, bounds=(1, None), doc="""
        Declares how many options are displayed at the same time.
        If set to 1 displays options as dropdown otherwise displays
        scrollable area (not currently supported).""")

    variant = param.Selector(objects=["filled", "outlined", "standard"], default="outlined")

    _constants = {"multi": False}
    _esm_base = "Select.jsx"
    _rename = {"name": "name", "groups": None}


class _RadioGroup(MaterialWidget):
    """
    Base class for Material UI radio groups.

    This class combines Material UI styling with Panel's radio group functionality.
    It provides the foundation for widgets that allow selecting a single value from
    a list of options, such as radio buttons and checkboxes.
    """

    color = param.Selector(default="primary", objects=COLORS)

    inline = param.Boolean(default=False, doc="""
        Whether the items be arrange vertically (``False``) or
        horizontally in-line (``True``).""")

    width = param.Integer(default=None)

    _esm_base = "RadioGroup.jsx"

    _rename = {"name": "name"}

    __abstract = True


class RadioBoxGroup(_RadioGroup, MaterialSingleSelectBase):
    """
    The `RadioBoxGroup` widget allows selecting a value from a list of options.

    It falls into the broad category of single-value, option-selection widgets
    that provide a compatible API and include the  `AutocompleteInput`,
    `Select` and `RadioButtonGroup` widgets.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/RadioBoxGroup.html
    - https://panel.holoviz.org/reference/widgets/RadioBoxGroup.html
    - https://mui.com/material-ui/react-radio-button/

    :Example:

    >>> RadioBoxGroup(
    ...     label='Study', options=['Biology', 'Chemistry', 'Physics'],
    ... )
    """

    orientation = param.Selector(default="horizontal", objects=["horizontal", "vertical"], doc="""
        Button group orientation, either 'horizontal' (default) or 'vertical'.""")

    value = param.Parameter(default=None, allow_None=True)

    _constants = {"exclusive": True}


class CheckBoxGroup(_RadioGroup, MaterialMultiSelectBase):
    """
    The `CheckBoxGroup` widget allows selecting between a list of options by
    ticking the corresponding checkboxes.

    It falls into the broad category of multi-option selection widgets that
    provide a compatible API that also include the `CheckButtonGroup` widget.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/CheckBoxGroup.html
    - https://panel.holoviz.org/reference/widgets/CheckBoxGroup.html
    - https://mui.com/material-ui/react-radio-button/

    :Example:

    >>> CheckBoxGroup(
    ...     name='Fruits', value=['Apple', 'Pear'], options=['Apple', 'Banana', 'Pear', 'Strawberry'],
    ... )
    """

    orientation = param.Selector(
        default="horizontal",
        objects=["horizontal", "vertical"],
        doc="""
        Button group orientation, either 'horizontal' (default) or 'vertical'.""",
    )

    value = param.List(default=None, allow_None=True)

    _constants = {"exclusive": False}


class _ButtonGroup(_ButtonLike):
    """
    Base class for Material UI button groups.

    This class combines Material UI styling with Panel's button group functionality.
    It provides the foundation for widgets that allow selecting a single or multiple
    values from a list of options, such as toggle buttons and checkboxes.

    This is an abstract base class and should not be used directly.
    """

    orientation = param.Selector(default="horizontal", objects=["horizontal", "vertical"], doc="""
        Button group orientation, either 'horizontal' (default) or 'vertical'.""")

    size = param.Selector(objects=["small", "medium", "large"], default="medium")

    width = param.Integer(default=None)

    _esm_base = "ButtonGroup.jsx"

    _rename = {"name": "name"}

    __abstract = True


class RadioButtonGroup(_ButtonGroup, MaterialSingleSelectBase):
    """
    The `RadioButtonGroup` widget allows selecting from a list or dictionary
    of values using a set of toggle buttons.

    It falls into the broad category of single-value, option-selection widgets
    that provide a compatible API and include the `AutocompleteInput`, `Select`,
    and `RadioBoxGroup` widgets.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/RadioButtonGroup.html
    - https://panel.holoviz.org/reference/widgets/RadioButtonGroup.html
    - https://mui.com/material-ui/react-toggle-button/

    :Example:

    >>> RadioButtonGroup(
    ...     label='Plotting library', options=['Matplotlib', 'Bokeh', 'Plotly'],
    ... )
    """

    value = param.Parameter()

    _constants = {"exclusive": True}


class CheckButtonGroup(_ButtonGroup, MaterialMultiSelectBase):
    """
    The `CheckButtonGroup` widget allows selecting from a list or dictionary
    of values using a set of toggle buttons.

    It falls into the broad category of multi-option selection widgets that
    provide a compatible API that also include the `CheckBoxGroup` widget.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/CheckButtonGroup.html
    - https://panel.holoviz.org/reference/widgets/CheckButtonGroup.html
    - https://mui.com/material-ui/react-toggle-button/

    :Example:

    >>> CheckButtonGroup(
    ...     label='Regression Models', value=['Lasso', 'Ridge'],
    ...     options=['Lasso', 'Linear', 'Ridge', 'Polynomial']
    ... )

    """

    _constants = {"exclusive": False}


class MultiSelect(MaterialMultiSelectBase):
    """
    The `MultiSelect` widget allows selecting multiple values from a list of
    `options`.

    It falls into the broad category of multi-value, option-selection widgets
    that provide a compatible API and include the `MultiSelect`,
    `CrossSelector`, `CheckBoxGroup` and `CheckButtonGroup` widgets.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/MultiSelect.html
    - https://panel.holoviz.org/reference/widgets/MultiSelect.html
    - https://mui.com/material-ui/react-select/#multiple-select

    >>> pmui.MultiSelect(label='MultiSelect', value=['Apple', 'Pear'],
    ...     options=['Apple', 'Banana', 'Pear', 'Strawberry'], size=8)
    """

    color = param.Selector(objects=COLORS, default="primary")

    max_items = param.Integer(default=None, bounds=(1, None), doc="""
        Maximum number of options that can be selected.""")

    size = param.Integer(default=None, doc="""
        The number of options to display at once (not currently supported).""")

    value = param.List(default=[], allow_None=True)

    variant = param.Selector(objects=["filled", "outlined", "standard"], default="outlined")

    _esm_base = "MultiSelect.jsx"


class MultiChoice(_SelectDropdownBase, MultiSelect):
    """
    The `MultiChoice` widget allows selecting multiple values from a list of
    `options`.

    It falls into the broad category of multi-value, option-selection widgets
    that provide a compatible API and include the `MultiSelect`,
    `CrossSelector`, `CheckBoxGroup` and `CheckButtonGroup` widgets.

    The `MultiChoice` widget provides a much more compact UI than
    `MultiSelect`.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/MultiChoice.html
    - https://panel.holoviz.org/reference/widgets/MultiChoice.html
    - https://mui.com/material-ui/react-select/#multiple-select

    :Example:

    >>> MultiChoice(
    ...     label='Favourites', value=['Panel', 'hvPlot'],
    ...     options=['Panel', 'hvPlot', 'HoloViews', 'GeoViews', 'Datashader', 'Param', 'Colorcet'],
    ...     max_items=2
    ... )
    """

    chip = param.Boolean(default=True, doc="Whether to display a chip for each selected option")

    delete_button = param.Boolean(default=True, doc="""
        Whether to display a button to delete a selected option.""")

    option_limit = param.Integer(default=None, bounds=(1, None), doc="""
        Maximum number of options to display at once.""")

    search_option_limit = param.Integer(default=None, bounds=(1, None), doc="""
        Maximum number of options to display at once if search string is entered.""")

    placeholder = param.String(default='', doc="""
        String displayed when no selection has been made.""")

    solid = param.Boolean(default=True, doc="""
        Whether to display chips with solid or outlined style.""")

    _constants = {"multi": True}
    _esm_base = "Select.jsx"
    _rename = {"name": None}


class NestedSelect(_PnNestedSelect):
    """
    The `NestedSelect` widget is composed of multiple widgets, where subsequent select options
    depend on the parent's value.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/NestedSelect.html
    - https://panel.holoviz.org/reference/widgets/NestedSelect.html
    - https://mui.com/material-ui/react-select/

    :Example:

    >>> NestedSelect(
    ...     options={
    ...         "gfs": {"tmp": [1000, 500], "pcp": [1000]},
    ...         "name": {"tmp": [1000, 925, 850, 700, 500], "pcp": [1000]},
    ...     },
    ...     levels=["model", "var", "level"],
    ... )
    """

    def _extract_level_metadata(self, i):
        """
        Extract the widget type and keyword arguments from the level metadata.
        """
        level = self._levels[i]
        if isinstance(level, int):
            return Select, {}
        elif isinstance(level, str):
            return Select, {"name": level}
        widget_type = level.get("type", Select)
        widget_kwargs = {k: v for k, v in level.items() if k != "type"}
        return widget_type, widget_kwargs


__all__ = [
    "AutocompleteInput",
    "Select",
    "RadioBoxGroup",
    "CheckBoxGroup",
    "RadioButtonGroup",
    "CheckButtonGroup",
    "MultiSelect",
    "MultiChoice",
    "NestedSelect",
]



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

import param
from panel._param import Margin
from panel.viewable import Viewable
from panel.widgets.base import WidgetBase

from ..base import ESMTransform, MaterialComponent

if TYPE_CHECKING:
    T = TypeVar('T')


class TooltipTransform(ESMTransform):

    _transform = """\
import Icon from "@mui/material/Icon";
import Tooltip from "@mui/material/Tooltip";

{esm}

function {output}(props) {{
  const [description] = props.model.useState("description")
  const [description_delay] = props.model.useState("description_delay")

  const widget = <{input} {{...props}} />
  return (description ? (
    <Tooltip
      title={{description}}
      arrow
      enterDelay={{description_delay}}
      enterNextDelay={{description_delay}}
      placement="right"
      slotProps={{{{ popper: {{ container: props.el }} }}}}
    >
      {{widget}}
    </Tooltip>) : widget
  )
}}
"""


class MaterialWidget(MaterialComponent, WidgetBase):
    """
    MaterialWidget is a base class for all Material UI widgets.
    """

    description = param.String(default=None)

    disabled = param.Boolean(default=False)

    label = param.String(default="")

    margin = Margin(default=10)

    width = param.Integer(default=300, bounds=(0, None), allow_None=True)

    __abstract = True

    def __init__(self, **params):
        if 'label' not in params and 'name' in params:
            params['label'] = params['name']
        super().__init__(**params)

    def _process_param_change(self, params):
        description = params.pop("description", None)
        icon = params.pop("icon", None)
        label = params.pop("label", None)
        props = MaterialComponent._process_param_change(self, params)
        if icon:
            props["icon"] = icon
        if label:
            props["label"] = label
        if description:
            props["description"] = description
        return props

    @classmethod
    def from_param(cls: type[T], parameter: param.Parameter, **params) -> T:
        """
        Construct a widget from a Parameter and link the two
        bi-directionally.

        Parameters
        ----------
        parameter: param.Parameter
          A parameter to create the widget from.

        Returns
        -------
        Widget instance linked to the supplied parameter
        """
        widget = super().from_param(parameter, **params)
        if isinstance(parameter.owner, MaterialComponent):
            widget.jslink(parameter.owner, value=parameter.name)
        return widget

    def api(self, jslink: bool=False, sizing_mode="stretch_width", **kwargs)->Viewable:
        """Returns an interactive component for exploring the API of the widget.

        Parameters
        ----------
        jslink: bool
            Whether to use jslinks instead of Python based links.
            This does not allow using all types of parameters.
        sizing_mode: str
            Sizing mode for the component.
        kwargs: dict
            Additional arguments to pass to the component.

        Example:
        --------
        >>> pmui.Button(name="Open").api()
        """
        import panel as pn

        import panel_material_ui as pmui
        return pmui.Tabs(
            pn.pane.HTML(self.param, name="Table", sizing_mode="stretch_width"),
            pmui.Row(self.controls(jslink=jslink), self, name="Editor", sizing_mode="stretch_width"),
            sizing_mode=sizing_mode, **kwargs
        )



import pytest

pytest.importorskip('playwright')

from bokeh.models.formatters import PrintfTickFormatter
from panel.tests.util import serve_component, wait_until
from panel_material_ui.widgets import IntSlider, Rating
from playwright.sync_api import expect

pytestmark = pytest.mark.ui


def test_int_slider(page):
    widget = IntSlider(value=5, start=0, end=10)
    serve_component(page, widget)

    slider = page.locator('.int-slider')
    expect(slider).to_have_count(1)

    slider_value = page.locator('.MuiFormLabel-root')
    expect(slider_value).to_have_text(str(widget.value))


def test_slider_value_update(page):
    widget = IntSlider(value=5, start=0, end=10)
    serve_component(page, widget)
    slider_value = page.locator('.MuiFormLabel-root')

    for i in range(widget.start, widget.end, widget.step):
        widget.value = i
        expect(slider_value).to_have_text(str(i))


@pytest.mark.parametrize('color', ['primary', 'secondary', 'error', 'info', 'success', 'warning'])
def test_slider_color(page, color):
    widget = IntSlider(value=5, start=0, end=10, color=color)
    serve_component(page, widget)

    expect(page.locator(f'.MuiSlider-color{color.capitalize()}')).to_have_count(1)


@pytest.mark.parametrize('track', ["inverted", False])
def test_slider_track(page, track):
    widget = IntSlider(value=5, start=0, end=10, track=track)
    serve_component(page, widget)
    expect(page.locator(f'.MuiSlider-track{str(track).capitalize()}')).to_have_count(1)


def test_slider_vertical_orientation(page):
    widget = IntSlider(value=5, start=0, end=10, orientation='vertical')

    serve_component(page, widget)

    expect(page.locator(f'.MuiSlider-vertical')).to_have_count(1)
    assert page.locator('.MuiSlider-rail').evaluate("el => el.offsetHeight") == 277


def test_slider_format_str(page):
    widget = IntSlider(value=1101, start=0, end=2000, format='0a')

    serve_component(page, widget)

    expect(page.locator('.MuiFormLabel-root')).to_have_text('1k')

    widget.value = 2000

    expect(page.locator('.MuiFormLabel-root')).to_have_text('2k')


def test_slider_format_model(page):
    widget = IntSlider(value=1, start=0, end=10, format=PrintfTickFormatter(format='%d m'))

    serve_component(page, widget)

    expect(page.locator('.MuiFormLabel-root')).to_have_text('1 m')

    widget.value = 7

    expect(page.locator('.MuiFormLabel-root')).to_have_text('7 m')


@pytest.mark.parametrize('size', ["small", "medium", "large"])
def test_rating(page, size):
    widget = Rating(value=3, size=size)
    serve_component(page, widget)

    rating = page.locator('.rating')
    expect(rating).to_have_count(1)

    rating_size = page.locator(f'.MuiRating-size{size.capitalize()}')
    expect(rating_size).to_have_count(1)



import pytest

pytest.importorskip('playwright')

from panel.tests.util import serve_component, wait_until
from panel_material_ui.widgets import AutocompleteInput, Select, RadioBoxGroup, RadioButtonGroup, CheckButtonGroup
from playwright.sync_api import expect

pytestmark = pytest.mark.ui


def test_autocomplete_input_value_updates(page):
    widget = AutocompleteInput(name='Autocomplete Input test', options=["Option 1", "Option 2", "123"])
    serve_component(page, widget)

    expect(page.locator(".autocomplete-input")).to_have_count(1)

    page.locator("input").fill("Option 2")
    page.locator(".MuiAutocomplete-option").click()

    wait_until(lambda: widget.value == 'Option 2', page)

def test_autocomplete_dict_options(page):
    widget = AutocompleteInput(name='Autocomplete Input test', options={"Option 1": 1, "Option 2": 2, "123": 123})
    serve_component(page, widget)

    expect(page.locator(".autocomplete-input")).to_have_count(1)

    page.locator("input").fill("Option 2")
    page.locator(".MuiAutocomplete-option").click()

    wait_until(lambda: widget.value == 2, page)

def test_autocomplete_input_value_updates_unrestricted(page):
    widget = AutocompleteInput(name='Autocomplete Input test', options=["Option 1", "Option 2", "123"], restrict=False)
    serve_component(page, widget)

    expect(page.locator(".autocomplete-input")).to_have_count(1)

    page.locator("input").fill("Option 3")
    page.locator("input").press("Enter")

    wait_until(lambda: widget.value == 'Option 3', page)

@pytest.mark.parametrize('variant', ["filled", "outlined", "standard"])
def test_autocomplete_input_variant(page, variant):
    widget = AutocompleteInput(name='Autocomplete Input test', variant=variant, options=["Option 1", "Option 2", "123"])
    serve_component(page, widget)

    expect(page.locator(".autocomplete-input")).to_have_count(1)
    expect(page.locator(f"div[variant='{variant}']")).to_have_count(1)

def test_autocomplete_input_search_strategy(page):
    widget = AutocompleteInput(name='Autocomplete Input test', options=["Option 1", "Option 2", "123"])
    serve_component(page, widget)

    expect(page.locator(".autocomplete-input")).to_have_count(1)

    page.locator("input").fill("Option")
    expect(page.locator(".MuiAutocomplete-option")).to_have_count(2)

    page.locator("input").fill("ti")
    expect(page.locator(".MuiAutocomplete-option")).to_have_count(0)

    widget.search_strategy = "includes"
    page.locator("input").fill("tion")
    expect(page.locator(".MuiAutocomplete-option")).to_have_count(2)

def test_autocomplete_input_case_sensitive(page):
    widget = AutocompleteInput(name='Autocomplete Input test', options=["Option 1", "Option 2", "123"])
    serve_component(page, widget)

    expect(page.locator(".autocomplete-input")).to_have_count(1)

    page.locator("input").fill("opt")
    expect(page.locator(".MuiAutocomplete-option")).to_have_count(0)

    widget.case_sensitive = False

    page.locator("input").fill("option")
    expect(page.locator(".MuiAutocomplete-option")).to_have_count(2)

def test_autocomplete_min_characters(page):
    widget = AutocompleteInput(name='Autocomplete Input test', options=["Option 1", "Option 2", "123"])
    serve_component(page, widget)

    expect(page.locator(".autocomplete-input")).to_have_count(1)

    page.locator("input").fill("O")
    expect(page.locator(".MuiAutocomplete-option")).to_have_count(0)
    page.locator("input").fill("")

    widget.min_characters = 1

    page.locator("input").fill("O")
    expect(page.locator(".MuiAutocomplete-option")).to_have_count(2)

@pytest.mark.parametrize('variant', ["filled", "outlined", "standard"])
def test_select_variant(page, variant):
    widget = Select(name='Select test', variant=variant, options=["Option 1", "Option 2", "Option 3"])
    serve_component(page, widget)

    expect(page.locator(".select")).to_have_count(1)
    expect(page.locator(f".MuiSelect-{variant}")).to_have_count(1)

def test_select_disabled_options(page):
    widget = Select(name='Select test', options=["Option 1", "Option 2", "Option 3"], disabled_options=["Option 2"])
    serve_component(page, widget)

    expect(page.locator(".select")).to_have_count(1)

    page.locator(".select").click(force=True)
    expect(page.locator(".MuiMenuItem-root")).to_have_count(3)
    expect(page.locator(".MuiMenuItem-root.Mui-disabled")).to_have_text("Option 2")

@pytest.mark.parametrize('color', ["primary", "secondary", "error", "info", "success", "warning"])
def test_radio_box_group_color(page, color):
    widget = RadioBoxGroup(name='RadioBoxGroup test', options=["Option 1", "Option 2", "Option 3"], color=color)
    serve_component(page, widget)

    expect(page.locator(".radio-box-group")).to_have_count(1)
    expect(page.locator(f".MuiRadio-color{color.capitalize()}")).to_have_count(len(widget.options))

@pytest.mark.parametrize('inline', [True, False])
def test_radio_box_group_orientation(page, inline):
    widget = RadioBoxGroup(name='RadioBoxGroup test', options=["Option 1", "Option 2", "Option 3"], inline=inline)
    serve_component(page, widget)

    expect(page.locator(".radio-box-group")).to_have_count(1)
    if inline:
        rbg_orient = page.locator(".MuiRadioGroup-row")
        expect(rbg_orient).to_have_count(1)

@pytest.mark.parametrize('button_type', ["primary", "secondary", "error", "info", "success", "warning"])
def test_radio_button_group_color(page, button_type):
    widget = RadioButtonGroup(
        name='RadioButtonGroup test',
        options=["Option 1", "Option 2", "Option 3"],
        button_type=button_type
    )
    serve_component(page, widget)

    expect(page.locator(".radio-button-group")).to_have_count(1)
    if button_type == "error":
        option_color = page.locator(f".Mui-{button_type}")
    else:
        option_color = page.locator(f".MuiToggleButton-{button_type}")
    expect(option_color).to_have_count(len(widget.options))

@pytest.mark.parametrize('orientation', ["horizontal", "vertical"])
def test_radio_button_group_orientation(page, orientation):
    widget = RadioButtonGroup(
        name='RadioButtonGroup test',
        options=["Option 1", "Option 2", "Option 3"],
        orientation=orientation
    )
    serve_component(page, widget)

    expect(page.locator(".radio-button-group")).to_have_count(1)
    expect(page.locator(f".MuiToggleButtonGroup-{orientation}")).to_have_count(1)

@pytest.mark.parametrize('size', ["small", "medium", "large"])
def test_radio_button_group_size(page, size):
    widget = RadioButtonGroup(
        name='RadioButtonGroup test',
        options=["Option 1", "Option 2", "Option 3"],
        size=size
    )
    serve_component(page, widget)

    expect(page.locator(".radio-button-group")).to_have_count(1)
    expect(page.locator(f".MuiToggleButton-size{size.capitalize()}")).to_have_count(len(widget.options))

@pytest.mark.parametrize('button_type', ["primary", "secondary", "error", "info", "success", "warning"])
def test_check_button_group_color(page, button_type):
    widget = CheckButtonGroup(
        name='CheckButtonGroup test',
        value=[],
        options=["Option 1", "Option 2", "Option 3"],
        button_type=button_type
    )
    serve_component(page, widget)

    expect(page.locator(".check-button-group")).to_have_count(1)
    if button_type == "error":
        option_color = page.locator(f".Mui-{button_type}")
    else:
        option_color = page.locator(f".MuiToggleButton-{button_type}")
    expect(option_color).to_have_count(len(widget.options))

@pytest.mark.parametrize('orientation', ["horizontal", "vertical"])
def test_check_button_group_orientation(page, orientation):
    widget = CheckButtonGroup(
        name='CheckButtonGroup test',
        value=[],
        options=["Option 1", "Option 2", "Option 3"],
        orientation=orientation
    )
    serve_component(page, widget)

    expect(page.locator(".check-button-group")).to_have_count(1)
    expect(page.locator(f".MuiToggleButtonGroup-{orientation}")).to_have_count(1)

@pytest.mark.parametrize('size', ["small", "medium", "large"])
def test_check_button_group_size(page, size):
    widget = CheckButtonGroup(
        name='CheckButtonGroup test',
        value=[],
        options=["Option 1", "Option 2", "Option 3"],
        size=size
    )
    serve_component(page, widget)

    expect(page.locator(".check-button-group")).to_have_count(1)
    expect(page.locator(f".MuiToggleButton-size{size.capitalize()}")).to_have_count(len(widget.options))



from __future__ import annotations

from collections.abc import Iterable
from datetime import date, datetime, timezone
from datetime import time as dt_time
from typing import TYPE_CHECKING, Any

import numpy as np
import param
from bokeh.models.formatters import NumeralTickFormatter, TickFormatter
from panel.models.reactive_html import DOMEvent
from panel.util import edit_readonly, try_datetime64_to_datetime
from panel.widgets.input import DatetimeInput as _PnDatetimeInput
from panel.widgets.input import FileInput as _PnFileInput
from panel.widgets.input import LiteralInput as _PnLiteralInput

from ..base import COLORS, LoadingTransform, ThemedTransform
from .base import MaterialWidget, TooltipTransform
from .button import _ButtonLike

if TYPE_CHECKING:
    from bokeh.document import Document


class MaterialInputWidget(MaterialWidget):

    color = param.Selector(objects=COLORS, default="default", doc="""
        The color variant of the input.""")

    variant = param.Selector(objects=["filled", "outlined", "standard"], default="outlined", doc="""
        The variant of the input.""")

    __abstract = True


class _TextInputBase(MaterialInputWidget):

    error_state = param.Boolean(
        default=False,
        doc="""
        Whether to display in error state.""",
    )

    max_length = param.Integer(
        default=5000,
        doc="""
        Max count of characters in the input field.""",
    )

    placeholder = param.String(
        default="",
        doc="""
        Placeholder for empty input field.""",
    )

    size = param.Selector(objects=["small", "medium", "large"], default="medium")

    value = param.String(default="")

    value_input = param.String(
        default="",
        allow_None=True,
        readonly=True,
        doc="""
        Initial or entered text value updated on every key press.""",
    )

    _constants = {"multiline": False}

    __abstract = True

    @param.depends('value', watch=True, on_init=True)
    def _sync_value_input(self):
        with edit_readonly(self):
            self.value_input = self.value


class TextInput(_TextInputBase):
    """
    The `TextInput` widget allows entering any string using a text input box.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/TextInput.html
    - https://panel.holoviz.org/reference/widgets/TextInput.html
    - https://mui.com/material-ui/react-text-field/

    :Example:

    >>> TextInput(label='Name', placeholder='Enter your name here ...')
    """

    enter_pressed = param.Event(doc="""
        Event when the enter key has been pressed.""")

    _esm_base = "TextField.jsx"

    def _handle_enter(self, event: DOMEvent):
        self.param.trigger('enter_pressed')


class PasswordInput(_TextInputBase):
    """
    The `PasswordInput` widget allows entering any string using an obfuscated text input box.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/PasswordInput.html
    - https://panel.holoviz.org/reference/widgets/PasswordInput.html
    - https://mui.com/material-ui/react-text-field/#input-adornments

    :Example:

    >>> PasswordInput(label='Password', placeholder='Enter your password here ...')
    """

    _esm_base = "PasswordField.jsx"


class TextAreaInput(_TextInputBase):
    """
    The `TextAreaInput` allows entering any multiline string using a text input
    box.

    Lines are joined with the newline character `\n`.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/TextAreaInput.html
    - https://panel.holoviz.org/reference/widgets/TextAreaInput.html
    - https://mui.com/material-ui/react-text-field/

    :Example:

    >>> TextAreaInput(
    ...     label='Description', placeholder='Enter your description here...'
    ... )
    """

    auto_grow = param.Boolean(
        default=False,
        doc="""
        Whether the text area should automatically grow vertically to
        accommodate the current text.""",
    )

    cols = param.Integer(
        default=20,
        doc="""
        Number of columns in the text input field.""",
    )

    max_rows = param.Integer(
        default=None,
        doc="""
        When combined with auto_grow this determines the maximum number
        of rows the input area can grow.""",
    )

    rows = param.Integer(
        default=2,
        doc="""
        Number of rows in the text input field.""",
    )

    resizable = param.ObjectSelector(
        objects=["both", "width", "height", False],
        doc="""
        Whether the layout is interactively resizable,
        and if so in which dimensions: `width`, `height`, or `both`.
        Can only be set during initialization.""",
    )

    _esm_base = "TextArea.jsx"


class FileInput(_ButtonLike, _PnFileInput):
    """
    The `FileInput` allows the user to upload one or more files to the server.

    It makes the filename, MIME type and (bytes) content available in Python.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/FileInput.html
    - https://panel.holoviz.org/reference/widgets/FileInput.html
    - https://mui.com/material-ui/react-button/#file-upload

    :Example:

    >>> FileInput(accept='.png,.jpeg', multiple=True)
    """

    width = param.Integer(default=None)

    _esm_base = "FileInput.jsx"
    _source_transforms = {
        'filename': None,
        'value': "'data:' + source.mime_type + ';base64,' + value"
    }

    def __init__(self, **params):
        super().__init__(**params)
        self._buffer = []

    def _handle_msg(self, msg: Any) -> None:
        status = msg["status"]
        if status == "in_progress":
            self._buffer.append(msg)
            return
        elif status == "initializing":
            return
        elif status == "finished":
            try:
                self._flush_buffer()
                self._send_msg({"status": "finished"})
            except Exception as e:
                self._send_msg({"status": "error", "error": str(e)})
        else:
            raise ValueError(f"Unknown status: {status}")

    def _flush_buffer(self):
        value, mime_type, filename = [], [], []
        for file_data in self._buffer:
            value.append(file_data["data"])
            filename.append(file_data["filename"])
            mime_type.append(file_data["mime_type"])
        if not (self.multiple or self.directory):
            value, filename, mime_type = value[0], filename[0], mime_type[0]
        self.param.update(
            filename=filename,
            mime_type=mime_type,
            value=value,
        )
        self._buffer.clear()

    def save(self, filename):
        """
        Saves the uploaded FileInput data object(s) to file(s) or
        BytesIO object(s).

        Parameters
        ----------
        filename (str or list[str]): File path or file-like object
        """
        _PnFileInput.save(self, filename)


class _NumericInputBase(MaterialInputWidget):

    format = param.ClassSelector(default=None, class_=(str, TickFormatter,), doc="""
        Allows defining a custom format string or bokeh TickFormatter.""")

    placeholder = param.String(default='0', doc="""
        Placeholder for empty input field.""")

    size = param.Selector(objects=["small", "medium", "large"], default="medium")

    start = param.Parameter(default=None, allow_None=True, doc="""
        Optional minimum allowable value.""")

    end = param.Parameter(default=None, allow_None=True, doc="""
        Optional maximum allowable value.""")

    value = param.Number(default=0, allow_None=True, doc="""
        The current value of the spinner.""")

    __abstract = True

    def __init__(self, **params):
        if 'value' not in params:
            value = params.get('start', self.value)
            if value is not None:
                params['value'] = value
        if 'value' in params and 'value_throttled' in self.param:
            params['value_throttled'] = params['value']
        super().__init__(**params)

    def _get_properties(self, doc: Document):
        props = super()._get_properties(doc)
        if props['data'].format is None:
            props['data'].format = NumeralTickFormatter(format='0,0' if self.mode == 'int' else '0,0.0[000]')
        return props

    def _process_param_change(self, params):
        if 'format' in params and isinstance(params['format'], str):
            params['format'] = NumeralTickFormatter(format=params['format'])
        return super()._process_param_change(params)


class _IntInputBase(_NumericInputBase):

    value = param.Integer(default=0, allow_None=True, doc="""
        The current value of the spinner.""")

    start = param.Integer(default=None, allow_None=True, doc="""
        Optional minimum allowable value.""")

    end = param.Integer(default=None, allow_None=True, doc="""
        Optional maximum allowable value.""")

    mode = param.String(default='int', constant=True, doc="""
        Define the type of number which can be enter in the input""")

    __abstract = True


class _FloatInputBase(_NumericInputBase):

    value = param.Number(default=0, allow_None=True, doc="""
        The current value of the spinner.""")

    start = param.Number(default=None, allow_None=True, doc="""
        Optional minimum allowable value.""")

    end = param.Number(default=None, allow_None=True, doc="""
        Optional maximum allowable value.""")

    mode = param.String(default='float', constant=True, doc="""
        Define the type of number which can be enter in the input""")

    __abstract = True


class _SpinnerBase(_NumericInputBase):

    page_step_multiplier = param.Integer(default=10, bounds=(0, None), doc="""
        Defines the multiplication factor applied to step when the page up
        and page down keys are pressed.""")

    wheel_wait = param.Integer(default=100, doc="""
        Defines the debounce time in ms before updating `value_throttled` when
        the mouse wheel is used to change the input.""")

    width = param.Integer(default=300, allow_None=True, doc="""
      Width of this component. If sizing_mode is set to stretch
      or scale mode this will merely be used as a suggestion.""")

    _esm_base = "NumberInput.jsx"

    __abstract = True


class IntInput(_SpinnerBase, _IntInputBase):
    """
    The `IntInput` allows selecting an integer value using a spinbox.

    It behaves like a slider except that lower and upper bounds are optional
    and a specific value can be entered. The value can be changed using the
    keyboard (up, down, page up, page down), mouse wheel and arrow buttons.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/IntInput.html
    - https://panel.holoviz.org/reference/widgets/IntInput.html
    - https://mui.com/material-ui/react-text-field/#input-adornments

    :Example:

    >>> IntInput(name='Value', value=100, start=0, end=1000, step=10)
    """

    step = param.Integer(default=1, doc="""
        The step size.""")

    value_throttled = param.Integer(default=None, constant=True, doc="""
        The current value. Updates only on `<enter>` or when the widget looses focus.""")


class FloatInput(_SpinnerBase, _FloatInputBase):
    """
    The `FloatInput` allows selecting an integer value using a spinbox.

    It behaves like a slider except that lower and upper bounds are optional
    and a specific value can be entered. The value can be changed using the
    keyboard (up, down, page up, page down), mouse wheel and arrow buttons.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/FloatInput.html
    - https://panel.holoviz.org/reference/widgets/FloatInput.html

    :Example:

    >>> FloatInput(label='Value', value=100, start=0, end=1000, step=10)
    """

    step = param.Number(default=0.1, doc="""
        The step size.""")

    value_throttled = param.Number(default=None, constant=True, doc="""
        The current value. Updates only on `<enter>` or when the widget looses focus.""")


class NumberInput(_SpinnerBase):

    def __new__(self, **params):
        param_list = ["value", "start", "stop", "step"]
        if all(isinstance(params.get(p, 0), int) for p in param_list):
            return IntInput(**params)
        else:
            return FloatInput(**params)


class _DatePickerBase(MaterialInputWidget):

    as_numpy_datetime64 = param.Boolean(default=None, doc="""
        Whether to return values as numpy.datetime64. If left unset,
        will be True if value is a numpy.datetime64, else False.""")

    clearable = param.Boolean(default=True, doc="If true, allows the date to be cleared.")

    disabled_dates = param.List(default=None, item_type=(date, str), doc="""
      Dates to make unavailable for selection.""")

    disable_future = param.Boolean(default=False, doc="If true, future dates are disabled.")

    disable_past = param.Boolean(default=False, doc="If true, past dates are disabled.")

    enabled_dates = param.List(default=None, item_type=(date, str), doc="""
      Dates to make available for selection.""")

    end = param.Date(default=None, doc="The maximum selectable date.")

    format = param.String(default='YYYY-MM-DD', doc="The format of the date displayed in the input.")

    open_to = param.Selector(objects=['year', 'month', 'day'], default='day', doc="The default view to open the calendar to.")

    show_today_button = param.Boolean(default=False, doc="If true, shows a button to select today's date.")

    start = param.Date(default=None, doc="The minimum selectable date.")

    value = param.Date(default=None, doc="The selected date.")

    views = param.List(default=['year', 'month', 'day'], doc="The views that are available for the date picker.")

    width = param.Integer(default=300, allow_None=True, doc="""
      Width of this component. If sizing_mode is set to stretch
      or scale mode this will merely be used as a suggestion.""")

    _esm_base = "DateTimePicker.jsx"

    _constants = {'range': False, 'time': False}

    __abstract = True

    def __init__(self, **params):
        # Since options is the standard for other widgets,
        # it makes sense to also support options here, converting
        # it to enabled_dates
        if 'options' in params:
            options = list(params.pop('options'))
            params['enabled_dates'] = options
        if 'value' in params:
            value = try_datetime64_to_datetime(params['value'])
            if hasattr(value, "date"):
                value = value.date()
            params["value"] = value
        super().__init__(**params)

    @staticmethod
    def _convert_date_to_string(v):
        return v.strftime('%Y-%m-%d')

    def _process_property_change(self, msg):
        msg = super()._process_property_change(msg)
        for p in ('start', 'end', 'value'):
            if p not in msg:
                continue
            value = msg[p]
            if isinstance(value, str):
                msg[p] = datetime.date(datetime.strptime(value, '%Y-%m-%d'))
            elif isinstance(value, float):
                msg[p] = datetime.fromtimestamp(value / 1000, tz=timezone.utc).date()
        return msg


class DatePicker(_DatePickerBase):
    """
    The `DatePicker` allows selecting a `date` value using a text box
    and a date-picking utility.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/DatePicker.html
    - https://panel.holoviz.org/reference/widgets/DatePicker.html
    - https://mui.com/x/react-date-pickers/

    :Example:

    >>> DatePicker(
    ...     value=date(2025,1,1),
    ...     start=date(2025,1,1), end=date(2025,12,31),
    ...     name='Date'
    ... )
    """

    _constants = {'range': False, 'time': False}

    value = param.ClassSelector(default=None, class_=(datetime, date, str), doc="""
        The current value. Can be a datetime object or a string in ISO format.""")

    def _serialize_value(self, value):
        """Convert value for sending to JavaScript."""
        if isinstance(value, str) and value:
            try:
                if self.as_numpy_datetime64:
                    return np.datetime64(value)
                else:
                    return self._parse_datetime_string(value)
            except ValueError:
                return None
        return value


class _DatetimePickerBase(_DatePickerBase):
    """Base class for DatetimePicker components."""

    enable_seconds = param.Boolean(default=True, doc="""
      Enable editing of the seconds in the widget.""")

    enable_time = param.Boolean(default=True, doc="""
      Enable editing of the time in the widget.""")

    format = param.String(default=None, doc="""
        Format to display the datetime. Use custom format string based on dayjs.
        If None, will be set automatically based on military_time setting.
        For 12-hour: 'YYYY-MM-DD hh:mm a'
        For 24-hour: 'YYYY-MM-DD HH:mm'
        Add ':ss' to include seconds.""")

    military_time = param.Boolean(default=True, doc="""
      Whether to display time in 24 hour format.""")

    open_to = param.Selector(objects=['year', 'month', 'day', 'hours', 'minutes'], default=None, doc="""
      The default view to open the calendar to.""")

    views = param.List(default=['year', 'month', 'day', 'hours', 'minutes'], doc="""
      The views that are available for the date picker.""")

    _constants = {'range': False, 'time': True}

    __abstract = True

    def __init__(self, **params):
        # Preserve original value
        original_value = None
        if 'value' in params:
            original_value = params['value']

        # Handle value as string - we want to preserve the datetime object
        if 'value' in params and isinstance(params['value'], str):
            params['value'] = self._parse_datetime_string(params['value'])

        super().__init__(**params)

        # Override any date-only conversion that might have happened in parent class
        if original_value is not None and isinstance(original_value, (datetime, str)) and hasattr(self, 'value'):
            if isinstance(original_value, str):
                self.value = self._parse_datetime_string(original_value)
            elif isinstance(original_value, datetime):
                self.value = original_value

        # Set default format based on military_time
        if self.format is None:
            self._update_format_from_settings()

    @param.depends('military_time', 'enable_seconds', watch=True)
    def _update_format_from_settings(self):
        """Update format based on clock and seconds settings."""
        if self.military_time:
            self.format = 'YYYY-MM-DD HH:mm' + (':ss' if self.enable_seconds else '')
        else:
            self.format = 'YYYY-MM-DD hh:mm' + (':ss' if self.enable_seconds else '') + ' a'

    @staticmethod
    def _parse_datetime_string(value_str):
        """Convert datetime string to datetime object."""
        if not value_str:
            return None

        formats = [
            '%Y-%m-%d %H:%M:%S',  # 2023-01-01 14:30:00
            '%Y-%m-%d %H:%M',     # 2023-01-01 14:30
            '%Y-%m-%dT%H:%M:%S',  # 2023-01-01T14:30:00
            '%Y-%m-%dT%H:%M'      # 2023-01-01T14:30
        ]

        for fmt in formats:
            try:
                return datetime.strptime(value_str, fmt)
            except ValueError:
                continue

        # If we can't parse it as datetime, try date only
        try:
            return datetime.strptime(value_str, '%Y-%m-%d')
        except ValueError as exc:
            raise ValueError(f"Could not parse '{value_str}' as a datetime") from exc

    def _convert_to_datetime(self, v):
        if v is None:
            return None

        if isinstance(v, Iterable) and not isinstance(v, str):
            container_type = type(v)
            return container_type(
                self._convert_to_datetime(vv)
                for vv in v
            )

        # Handle string conversion
        if isinstance(v, str):
            return self._parse_datetime_string(v)

        # Handle other types
        v = try_datetime64_to_datetime(v)
        if isinstance(v, datetime):
            return v
        elif isinstance(v, date):
            # Convert date to datetime
            return datetime(v.year, v.month, v.day)
        else:
            raise ValueError(f"Could not convert {v} to datetime")

    def _process_property_change(self, msg):
        """Process incoming values from JavaScript.

        IMPORTANT: We completely override the parent method to avoid
        the date-only parsing that causes problems with datetime strings.
        """
        # Get the parent's parent method (MaterialWidget._process_property_change)
        # to avoid _DatePickerBase's implementation which only handles dates
        msg = MaterialWidget._process_property_change(self, msg)

        # Handle our datetime parameters
        for p in ('start', 'end', 'value'):
            if p not in msg:
                continue

            value = msg[p]
            if isinstance(value, str):
                try:
                    # Parse as datetime with time component
                    msg[p] = self._parse_datetime_string(value)
                except ValueError:
                    # If parsing fails, keep the original value
                    pass

        return msg

    def _process_param_change(self, msg):
        """Process outgoing values to JavaScript."""
        # Only handle our specific parameters, let parent class handle the rest
        # This avoids the date-only conversion in the parent class
        our_params = {}
        for p in ('value', 'start', 'end'):
            if p in msg:
                our_params[p] = msg.pop(p)

        # Let parent handle the remaining parameters
        msg = super()._process_param_change(msg)

        # Now handle our parameters
        for p, v in our_params.items():
            if v is not None:
                # Convert to datetime if needed
                dt_value = self._convert_to_datetime(v)
                if dt_value is not None:
                    # Format as string for JavaScript
                    msg[p] = dt_value.strftime('%Y-%m-%d %H:%M:%S')
                else:
                    msg[p] = None

        return msg


class DatetimePicker(_DatetimePickerBase):
    """
    The `DatetimePicker` allows selecting selecting a `datetime` value using a
    textbox and a datetime-picking utility.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/DatetimePicker.html
    - https://panel.holoviz.org/reference/widgets/DatetimePicker.html
    - https://mui.com/x/react-date-pickers/

    :Example:

    >>> DatetimePicker(
    ...    value=datetime(2025,1,1,22,0),
    ...    start=date(2025,1,1), end=date(2025,12,31),
    ...    military_time=True, name='Date and time'
    ... )

    Also supports string values:

    >>> DatetimePicker(
    ...    value="2025-01-01 22:00:00",
    ...    military_time=True, name='Date and time'
    ... )
    """

    value = param.ClassSelector(default=None, class_=(datetime, date, str), doc="""
        The current value. Can be a datetime object or a string in ISO format.""")

    _source_transforms = {
        "value": None, "start": None, "end": None
    }

    def _serialize_value(self, value):
        """Convert value for sending to JavaScript."""
        if isinstance(value, str) and value:
            try:
                if self.as_numpy_datetime64:
                    return np.datetime64(value)
                else:
                    return self._parse_datetime_string(value)
            except ValueError:
                return None
        return value



class _TimeCommon(MaterialWidget):

    clock = param.Selector(default='12h', objects=['12h', '24h'], doc="""
        Whether to use 12 hour or 24 hour clock.""")

    hour_increment = param.Integer(default=1, bounds=(1, None), doc="""
        Defines the granularity of hour value increments in the UI.""")

    minute_increment = param.Integer(default=1, bounds=(1, None), doc="""
        Defines the granularity of minute value increments in the UI.""")

    second_increment = param.Integer(default=1, bounds=(1, None), doc="""
        Defines the granularity of second value increments in the UI.""")

    seconds = param.Boolean(default=False, doc="""
        Allows to select seconds. By default only hours and minutes are
        selectable, and AM/PM depending on the `clock` option.""")

    __abstract = True


class TimePicker(_TimeCommon):
    """
    The `TimePicker` allows selecting a `time` value using a text box
    and a time-picking utility.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/TimePicker.html
    - https://panel.holoviz.org/reference/widgets/TimePicker.html
    - https://mui.com/x/react-date-pickers/time-picker/

    :Example:

    >>> TimePicker(
    ...     value=time(12, 59, 31), start="09:00:00", end="18:00:00", label="Time"
    ... )
    """

    color = param.Selector(objects=COLORS, default="default")

    value = param.ClassSelector(default=None, class_=(dt_time, str), doc="""
        The current value""")

    start = param.ClassSelector(default=None, class_=(dt_time, str), doc="""
        Inclusive lower bound of the allowed time selection""")

    end = param.ClassSelector(default=None, class_=(dt_time, str), doc="""
        Inclusive upper bound of the allowed time selection""")

    format = param.String(default=None, doc="""
        Format to display the time. Use 'HH:mm:ss' to include seconds.
        For 12-hour clock, use 'hh:mm a'. See dayjs formatting options.
        If None, will be automatically set based on clock and seconds settings.

        +----+------------------------------------+------------+
        | H  | Hours                              | 0 to 23    |
        | HH | Hours, 2-digits                    | 00 to 23   |
        | h  | Hours, 12-hour clock               | 1 to 12    |
        | hh | Hours, 12-hour clock, 2-digits     | 1 to 12    |
        | m  | Minutes                            | 0 to 59    |
        | mm | Minutes                            | 00 to 59   |
        | s  | Seconds                            | 0, 1 to 59 |
        | ss | Seconds                            | 00 to 59   |
        | a  | am/pm, lower-case                  | am or pm   |
        | A  | AM/PM, upper-cas                   | AM or PM   |
        +----+------------------------------------+------------+

    """)

    variant = param.Selector(objects=["filled", "outlined", "standard"], default="outlined")

    _esm_base = "TimePicker.jsx"

    def __init__(self, **params):
        # Convert time objects to appropriate string format if needed
        for attr in ['value', 'start', 'end']:
            if attr in params and isinstance(params[attr], dt_time):
                params[attr] = params[attr].strftime('%H:%M:%S')

        super().__init__(**params)

        # Set initial format based on clock and seconds if not explicitly provided
        if self.format is None:
            self._update_format_from_settings()

    @param.depends('clock', 'seconds', watch=True)
    def _update_format_from_settings(self):
        """Update the time format based on clock and seconds settings."""
        if self.clock == '12h':
            self.format = 'hh:mm a' if not self.seconds else 'hh:mm:ss a'
        else:  # 24h
            self.format = 'HH:mm' if not self.seconds else 'HH:mm:ss'

    def _process_param_change(self, msg):
        msg = super()._process_param_change(msg)
        if 'value' in msg and isinstance(msg['value'], dt_time):
            msg['value'] = msg['value'].strftime('%H:%M:%S')
        if 'start' in msg and isinstance(msg['start'], dt_time):
            msg['start'] = msg['start'].strftime('%H:%M:%S')
        if 'end' in msg and isinstance(msg['end'], dt_time):
            msg['end'] = msg['end'].strftime('%H:%M:%S')
        if 'format' in msg and msg['format'] is not None:
            msg['format'] = msg['format'].replace('K', 'A').replace('G', 'HH').replace('i', 'mm')
        return msg

    def _process_property_change(self, msg):
        msg = super()._process_property_change(msg)
        # Convert string time values to datetime.time objects
        if 'value' in msg:
            time_str = msg['value']
            if time_str and isinstance(time_str, str):
                time_parts = time_str.split(':')
                if len(time_parts) >= 2:
                    hours = int(time_parts[0])
                    minutes = int(time_parts[1])
                    seconds = int(time_parts[2]) if len(time_parts) > 2 else 0
                    msg['value'] = dt_time(hours, minutes, seconds)
        return msg


class Checkbox(MaterialWidget):
    """
    The `Checkbox` allows toggling a single condition between `True`/`False`
    states by ticking a checkbox.

    This widget is interchangeable with the `Switch` widget.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/Checkbox.html
    - https://panel.holoviz.org/reference/widgets/Checkbox.html
    - https://mui.com/material-ui/react-checkbox/

    :Example:

    >>> Checkbox(label='Works with the tools you know and love', value=True)
    """

    color = param.Selector(objects=COLORS, default="default")

    description_delay = param.Integer(default=1000, doc="""
        Delay (in milliseconds) to display the tooltip after the cursor has
        hovered over the Button, default is 1000ms.""")

    indeterminate = param.Boolean(default=False)

    size = param.Selector(objects=["small", "medium", "large"], default="medium")

    value = param.Boolean(default=False)

    width = param.Integer(default=None)

    _esm_base = "Checkbox.jsx"
    _esm_transforms = [TooltipTransform, ThemedTransform]


class Switch(MaterialWidget):
    """
    The `Switch` allows toggling a single condition between `True`/`False`
    states by ticking a checkbox.

    This widget is interchangeable with the `Checkbox` widget.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/Switch.html
    - https://panel.holoviz.org/reference/widgets/Switch.html
    - https://mui.com/material-ui/react-switch/

    :Example:

    >>> Switch(label='Works with the tools you know and love', value=True)
    """

    color = param.Selector(objects=COLORS, default="default")

    description_delay = param.Integer(default=1000, doc="""
        Delay (in milliseconds) to display the tooltip after the cursor has
        hovered over the Button, default is 1000ms.""")

    edge = param.Selector(objects=["start", "end", False], default=False)

    size = param.Selector(objects=["small", "medium", "large"], default="medium")

    value = param.Boolean(default=False)

    width = param.Boolean(default=None)

    _esm_base = "Switch.jsx"
    _esm_transforms = [LoadingTransform, TooltipTransform, ThemedTransform]


class ColorPicker(MaterialWidget):
    """
    The `ColorPicker` allows selecting a color value using a color picker utility.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/ColorPicker.html
    - https://panel.holoviz.org/reference/widgets/ColorPicker.html
    - https://viclafouch.github.io/mui-color-input/

    :Example:

    >>> pmui.ColorPicker(name='Color Picker', value='#99ef78')
    """

    alpha = param.Boolean(default=False, doc="Whether to allow alpha transparency.")

    color = param.Selector(objects=COLORS, default="default")

    format = param.Selector(objects=["hex", "rgb", "rgba", "hsl", "hsv"], default="hex")

    size = param.Selector(objects=["small", "medium", "large"], default="medium")

    variant = param.Selector(objects=["filled", "outlined", "standard"], default="outlined")

    value = param.String(default=None, doc="The current color value.")

    _esm_base = "ColorPicker.jsx"


class LiteralInput(TextInput, _PnLiteralInput):
    """
    The `LiteralInput` allows entering any string using a text input box.

    :References:

    - https://panel.holoviz.org/reference/widgets/LiteralInput.html
    """

    value = param.Parameter()

    value_input = param.Parameter()

    def _process_param_change(self, msg):
        msg = super()._process_param_change(msg)
        if self._state:
            msg["error_state"] = True
        else:
            msg["error_state"] = False
        if "value" in msg:
            msg["value_input"] = msg.pop("value")
        if "title" in msg:
            msg["label"] = msg.pop("title")
        return msg


class DatetimeInput(TextInput, _PnDatetimeInput):
    """
    The `DatetimeInput` allows entering a datetime value using a text input box.
    """

    value = param.ClassSelector(default=None, class_=(datetime, date, str), doc="""
        The current value. Can be a datetime object or a string in ISO format.""")

    value_input = param.ClassSelector(default=None, class_=(datetime, date, str), doc="""
        The current value. Can be a datetime object or a string in ISO format.""")

    _source_transforms = {
        "value": None, "value_input": None, "start": None, "end": None
    }

    def _process_param_change(self, msg):
        msg = super()._process_param_change(msg)
        if self._state:
            msg["error_state"] = True
        else:
            msg["error_state"] = False
        if "value" in msg:
            msg["value_input"] = msg.pop("value")
        if "title" in msg:
            msg["label"] = msg.pop("title")
        return msg


class DictInput(LiteralInput):
    """
    The `DictInput` allows entering a dictionary value using a text input box.
    """

    type = param.ClassSelector(default=dict, class_=type, readonly=True, doc="The type of the value.")

    value = param.Parameter(default={})


class ListInput(LiteralInput):
    """
    The `ListInput` allows entering a list value using a text input box.
    """

    type = param.ClassSelector(default=list, class_=type, readonly=True, doc="The type of the value.")

    value = param.Parameter(default=[])


class TupleInput(LiteralInput):
    """
    The `TupleInput` allows entering a tuple value using a text input box.
    """

    type = param.ClassSelector(default=tuple, class_=type, readonly=True, doc="The type of the value.")

    value = param.Parameter(default=())


__all__ = [
    "TextInput",
    "PasswordInput",
    "TextAreaInput",
    "FileInput",
    "IntInput",
    "FloatInput",
    "NumberInput",
    "DatePicker",
    "DatetimePicker",
    "TimePicker",
    "Checkbox",
    "Switch",
    "ColorPicker",
    "LiteralInput",
    "DatetimeInput",
    "DictInput",
    "ListInput",
    "TupleInput"
]



from copy import deepcopy

import colorsys
import holoviews as hv
import hvplot.pandas
import numpy as np
import panel as pn
import panel_material_ui as pmu
import pandas as pd

pn.extension(sizing_mode="stretch_width")

pmu.Paper.margin = 10

LORUM_IPSUM = """\
**Lorem Ipsum** is simply dummy text of the printing and typesetting industry. Lorem Ipsum has been the
industry's standard dummy text ever since the 1500s, when an unknown printer took a galley of type
and scrambled it to make a type specimen book"""

# App Code

@pn.cache
def get_data(n_categories: int, n_rows=100):
    x = np.random.rand(n_rows)
    y = np.random.rand(n_rows)

    categories = [f"Category {i+1}" for i in range(n_categories)]
    category = np.random.choice(categories, n_rows)

    dataframe = pd.DataFrame({"x": x, "y": y, "category": category}).sort_values('category')

    return dataframe

def get_categorical_plot(df, palette):
    return df.hvplot.scatter(
        x="x",
        y="y",
        size=75,
        color="category",
        cmap=palette,
        height=350,
        responsive=True,
        title="Categorical Plot",
        tools=["fullscreen"],
        legend='top_right',
    ).opts(
        backend_opts={
            'plot.toolbar.autohide': True
        },
        toolbar='above'
    )

def get_continous_plot(color="#9c27b0"):
    np.random.seed(42)
    data = pd.DataFrame(
        {
            "x": np.random.rand(100),
            "y": np.random.rand(100),
            "value": np.random.rand(100) * 100,
        }
    )
    custom_cmap = pmu.theme.linear_gradient("#ffffff", color, n=256)
    return data.hvplot.scatter(
        x="x",
        y="y",
        c="value",
        size=75,
        cmap=custom_cmap,
        colorbar=True,
        height=350,
        responsive=True,
        title="Continuous Plot",
        tools=['fullscreen'],
    ).opts(
        backend_opts={
            'plot.toolbar.autohide': True
        },
        toolbar='above'
    )

# Configure theme
paper = pmu.Checkbox(
    value=True, label="Paper", sizing_mode="fixed", align="center"
)
font_family = pmu.Select(
    label="Font", options=["Roboto", "Impact", "Palatino Linotype"]
)
primary_color = pmu.ColorPicker(
    value="#9c27b0",
    name="Primary Color",
)

theme_config = {
    "palette": {
        "primary": {
            "main": primary_color,
        },
    },
    "typography": {
        "fontFamily": (font_family,),
    },
}

n_colors = pmu.IntSlider(
    value=3,
    start=1,
    end=10,
    name="Categories",
)

theme = pmu.ThemeToggle(sizing_mode="fixed", align="center")
action_row = pmu.Row(paper, primary_color, n_colors, font_family, pn.HSpacer(), theme)


# Set up  plots
data = pn.bind(get_data, n_colors)
palette = pn.bind(pmu.theme.generate_palette, primary_color, n_colors)

categorical_plot = hv.DynamicMap(pn.bind(
    get_categorical_plot,
    df=data,
    palette=palette
))

continous_plot = hv.DynamicMap(pn.bind(
    get_continous_plot, color=primary_color
))


colors_out = pmu.TextInput(
    value=pn.rx(', ').join(palette),
    name="Categorical Colors",
    disabled=True,
)
button_out = pmu.Button(label="Click Me", color="primary")
column_out = pmu.Column(colors_out, button_out, LORUM_IPSUM)

elevation = paper.rx().rx.where(1, 0)

pmu.Container(
    pmu.Paper(action_row, elevation=elevation),
    pmu.Paper(continous_plot, elevation=elevation),
    pmu.Paper(categorical_plot, elevation=elevation),
    pmu.Paper(column_out, elevation=elevation),
    theme_config=theme_config, width_option='md'
).servable()



from __future__ import annotations

import datetime
from contextlib import ExitStack
from io import BytesIO
from pathlib import PurePath
from zoneinfo import ZoneInfo

import param
from panel.chat.message import DEFAULT_AVATARS as DEFAULT_AVATARS_PANEL
from panel.chat.message import ChatMessage, ChatReactionIcons
from panel.io import state
from panel.layout import Panel, Row
from panel.pane import Placeholder
from panel.pane import panel as as_panel
from panel.pane.image import FileBase, Image, ImageBase
from panel.pane.markup import HTMLBasePane
from panel.util import isfile
from panel.viewable import Child
from panel.widgets import Widget

from ..base import MaterialComponent
from .input import ChatAreaInput

_MESSAGE_BG = ":host(.message), .message { background-color: unset !important; box-shadow: unset !important; }"

DEFAULT_AVATARS = {
    "system": {"type": "icon", "icon": "settings"},
    **DEFAULT_AVATARS_PANEL
}


class MessageState(param.Parameterized):

    avatar = param.Parameter(allow_refs=True)

    timestamp = param.String(allow_refs=True)


class ChatMessage(MaterialComponent, ChatMessage):
    """
    Renders another component as a chat message with an associated user
    and avatar with support for various content types.

    This widget provides a structured view of chat messages, including features like:

    - Displaying user avatars, which can be text, emoji, or images.
    - Showing the user's name.
    - Displaying the message timestamp in a customizable format.
    - Associating reactions with messages and mapping them to icons.
    - Rendering various content types including text, images, audio, video, and more.

    :References:

    - https://panel-material-ui.holoviz.org/reference/chat/ChatMessage.html
    - https://panel.holoviz.org/reference/chat/ChatMessage.html

    :Example:

    >>> ChatMessage(object="Hello world!", user="New User", avatar="😊")
    """

    avatar = param.ClassSelector(default="", class_=(str, BytesIO, bytes, ImageBase, dict), doc="""
        The avatar to use for the user. Can be a single character text, an emoji, or anything
        supported by `pn.pane.Image`. If not set, checks if the user is available in the
        default_avatars mapping; else uses the first character of the name.""")

    css_classes = param.List(default=[],doc="""
        The CSS classes to apply to the widget.""")

    default_avatars = param.Dict(default=DEFAULT_AVATARS, doc="""
        A default mapping of user names to their corresponding avatars
        to use when the user is specified but the avatar is. You can
        modify, but not replace the dictionary.""")

    default_layout = param.ClassSelector(class_=(Panel), precedence=-1)

    elevation = param.Integer(default=2, doc="The elevation of the message.")

    placement = param.Selector(default="left", objects=["left", "right"], doc="The placement of the message.")

    _internal_state = param.ClassSelector(class_=MessageState, default=MessageState())
    _object_panel = Child()

    _esm_base = "ChatMessage.jsx"
    _rename = {
        "avatar": None,
        "avatar_lookup": None,
        "default_avatars": None,
        "object": None
    }

    def __init__(self, object=None, **params):
        self._exit_stack = ExitStack()
        if 'placement' not in params and ChatMessage.placement is None:
            user = params.get('user', ChatMessage.user).lower()
            params['placement'] = 'right' if user == 'user' else 'left'
        if params.get("timestamp") is None:
            tz = params.get("timestamp_tz")
            if tz is not None:
                tz = ZoneInfo(tz)
            elif state.browser_info and state.browser_info.timezone:
                tz = ZoneInfo(state.browser_info.timezone)
            params["timestamp"] = datetime.datetime.now(tz=tz)
        reaction_icons = params.get("reaction_icons", {"favorite": "heart"})
        if isinstance(reaction_icons, dict):
            params["reaction_icons"] = ChatReactionIcons(options=reaction_icons, default_layout=Row, sizing_mode=None)
        self._internal = True
        if not self.avatar:
            self._update_avatar()
        self._internal_state = MessageState(
            avatar=self.param.avatar,
        )
        if not ChatMessage.width and params.get('width') is None and params.get('sizing_mode', None) is None:
            params['sizing_mode'] = 'stretch_width'
        MaterialComponent.__init__(self, object=object, **params)
        self._internal_state.timestamp = self.param.timestamp.rx().strftime(self.param.timestamp_format)
        self._build_layout()

    @param.depends('avatar', watch=True, on_init=True)
    def _render_avatar_html(self):
        avatar = self.avatar
        if isinstance(avatar, dict):
            self._internal_state.avatar = avatar
        elif isinstance(avatar, ImageBase) or (isinstance(avatar, str) and Image.applies(avatar)):
            avatar = as_panel(avatar)
            if self.embed or (isfile(avatar.object) or not isinstance(avatar.object, (str, PurePath))):
                data = avatar._data(avatar.object)
                src = avatar._b64(data)
            elif isinstance(avatar.object, PurePath):
                raise ValueError(f"Could not find Avatar {type(avatar).__name__}.object {avatar.object}.")
            else:
                src = self.avatar.object
            self._internal_state.avatar = {"type": "image", "src": src}
        else:
            self._internal_state.avatar = {"type": "text", "text": self.avatar}

    @property
    def _synced_params(self) -> list[str]:
        """
        Parameters which are synced with properties using transforms
        applied in the _process_param_change method.
        """
        ignored = ['default_layout', 'loading', 'background']
        return [p for p in self.param if p not in self._manual_params+ignored]

    def _handle_msg(self, msg):
        if msg == 'edit':
            self._toggle_edit(msg)
        elif msg == 'copy':
            object_panel = self._object_panel
            if isinstance(object_panel, HTMLBasePane):
                object_panel = object_panel.object
            elif isinstance(object_panel, Widget):
                object_panel = object_panel.value
            self._send_msg({"type": "copy", "text": object_panel if isinstance(object_panel, str) else ""})

    def _build_layout(self):
        self._object_panel = self._create_panel(self.object)
        self._placeholder = Placeholder(
            object=self._object_panel,
            css_classes=["placeholder"],
            stylesheets=self._stylesheets + self.param.stylesheets.rx(),
            sizing_mode='stretch_width',
        )
        self._edit_area = ChatAreaInput(
            css_classes=["edit-area"],
            stylesheets=self._stylesheets + self.param.stylesheets.rx()
        )
        self.param.watch(self._update_object_pane, "object")
        self.param.watch(self._update_reaction_icons, "reaction_icons")
        self.edit_icon.param.watch(self._toggle_edit, "value")
        self._edit_area.param.watch(self._submit_edit, "enter_pressed")
        self._composite = Row()

    def _include_styles(self, obj):
        obj = as_panel(obj)
        combined = self._stylesheets + self.stylesheets + [_MESSAGE_BG]
        for o in obj.select():
            params = {
                "stylesheets": [
                    stylesheet for stylesheet in combined
                    if stylesheet not in o.stylesheets
                ] + o.stylesheets
            }
            is_markup = isinstance(o, HTMLBasePane) and not isinstance(o, FileBase)
            if is_markup:
                params["sizing_mode"] = None
                if not o.css_classes and len(str(o.object)) > 0:  # only show a background if there is content
                    params["css_classes"] = [
                        *(css for css in o.css_classes if css != "message"), "message"
                    ]
            o.param.update(**params)

    def _process_param_change(self, params):
        params = super()._process_param_change(params)
        if 'stylesheets' in params:
            params['stylesheets'] += [".message { font-size: 1.25em; overflow-wrap: anywhere; word-break: break-word; width; max-width: calc(100% - 20px); }"]
        return params

__all__ = ["ChatMessage"]



from __future__ import annotations

from typing import Any

import param
from panel.pane.markup import Markdown

from ..base import COLORS, MaterialComponent


class MaterialPaneBase(MaterialComponent):

    object = param.Parameter()

    __abstract = True

    def __init__(self, object=None, **params):
        super().__init__(object=object, **params)


class Avatar(MaterialPaneBase):
    """
    The `Avatar` component is used to display profile pictures, user initials, icons,
    or custom images.

    :References:

    - https://mui.com/material-ui/react-avatar/

    :Example:
    >>> Avatar("path/to/image.jpg")
    """

    alt_text = param.String(
        default=None,
        doc="""
        alt text to add to the image tag. The alt text is shown when a
        user cannot load or display the image.""",
    )

    color = param.Color()

    object = param.String(default="")

    size = param.Selector(objects=["small", "medium"], default="medium")

    variant = param.Selector(objects=["rounded", "square"], default="rounded")

    _esm_base = "Avatar.jsx"


class Chip(MaterialPaneBase):
    """
    A `Chip` can be used to display information, labels, tags, or actions. It can include text,
    an avatar, an icon, or a delete button.

    :References:

    - https://mui.com/material-ui/react-chip/

    :Example:
    >>> Chip("Log Time", icon="clock")
    """

    color = param.Selector(objects=COLORS, default="primary")

    icon = param.String(
        default=None,
        doc="""
        The name of the icon to display.""",
    )

    object = param.String(default="")

    size = param.Selector(objects=["small", "medium"], default="medium")

    variant = param.Selector(objects=["filled", "outlined"], default="filled")

    _esm_base = "Chip.jsx"

    def _handle_click(self, event):
        pass


class Skeleton(MaterialPaneBase):
    """
    The `Skeleton` component is used as a placeholder while content is loading.
    It provides a visual indication that data is being fetched, improving perceived performance
    and user experience.

    :References:

    - https://mui.com/material-ui/react-skeleton/
    """

    variant = param.Selector(objects=["circular", "rectangular", "rounded"], default="rounded")

    height = param.Integer(default=0)

    width = param.Integer(default=0)

    _esm_base = "Skeleton.jsx"


class Typography(MaterialPaneBase, Markdown):
    """
    The `Typography` component is used to display text with different styles and weights.

    :References:

    - https://mui.com/material-ui/react-typography/
    """

    variant = param.String(default=None, doc="The typography variant, e.g. h1, h2, body1.")

    _esm_base = "Typography.jsx"
    _rename = {"object": "object"}

    @classmethod
    def applies(cls, obj: Any) -> float | bool | None:
        if hasattr(obj, '_repr_markdown_'):
            return 0.29
        elif isinstance(obj, str):
            return 0.09
        else:
            return False



from __future__ import annotations

from collections import defaultdict
from functools import partial
from typing import Callable

import param
from panel._param import Margin
from panel.io.state import state
from panel.layout import Column
from panel.layout.base import ListLike
from panel.models.reactive_html import DOMEvent

from ..base import COLORS
from .base import MaterialWidget
from .button import _ButtonBase


def filter_item(item, keys):
    if isinstance(item, dict):
        item = {k: v for k, v in item.items() if k in keys}
        if 'children' in item:
            item['children'] = [filter_item(child, keys) for child in item['children']]
        return item
    return item


class MenuBase(MaterialWidget):

    active = param.Integer(default=None, doc="""
        The index of the currently selected menu item.""")

    items = param.ClassSelector(default=[], class_=list, doc="""
        List of items to display. Each item may be a string, a tuple mapping from a label to a value,
        or an object with a few common properties and a few widget specific properties.""")

    margin = Margin(default=0)

    value = param.ClassSelector(default=None, class_=(dict, str), doc="""
        Last clicked menu item.""")

    width = param.Integer(default=None, doc="""
        The width of the menu.""")

    _item_keys = ['label', 'children']
    _rename = {'value': None}
    _source_transforms = {'attached': None, "value": None, 'items': None}

    __abstract = True

    def __init__(self, **params):
        click_handler = params.pop('on_click', None)
        super().__init__(**params)
        self._on_action_callbacks = defaultdict(list)
        self._on_click_callbacks = []
        if click_handler:
            self.on_click(click_handler)

    def _process_param_change(self, params):
        params = super()._process_param_change(params)
        if 'items' in params:
            if isinstance(params['items'], list) and any(isinstance(item, tuple) for item in params['items']):
                # Legacy format from Panel
                items = []
                for item in params['items']:
                    if isinstance(item, tuple):
                        items.append({"label": item[0], "value": item[1]})
                    else:
                        items.append(item)
            else:
                items = params['items']
            params['items'] = [filter_item(item, self._item_keys) for item in items]
        return params

    @param.depends('items', watch=True, on_init=True)
    def _sync_items(self):
         self.param.active.bounds = (0, len(self.items)-1)

    @param.depends('active', watch=True, on_init=True)
    def _sync_active(self):
         self.value = self._lookup_item(self.active)

    @param.depends('value', watch=True, on_init=True)
    def _sync_value(self):
        index = self._lookup_active_by_value(self.value)
        if index is None:
            self.active = None
        else:
            self.active = index if len(index) > 1 else index[0]

    def _lookup_active_by_value(self, item):
        if not self.items:
            return None
        queue = [([], 0, self.items)]
        while queue:
            path, depth, items = queue.pop(0)

            for i, current in enumerate(items):
                current_path = path + [i]
                if current == item:
                    return tuple(current_path)
                if isinstance(current, dict):
                    if current == item:
                        return tuple(current_path)
                    if 'items' in current:
                        queue.append((current_path, depth + 1, current['items']))
        return None

    def _lookup_item(self, index):
        if index is None:
            return
        indexes = index if isinstance(index, tuple) else [index]
        value = self.items
        for i, idx in enumerate(indexes):
            value = value[idx]
            if isinstance(value, dict) and (i != len(indexes)-1):
                value = value['items']

        if isinstance(value, tuple):
            value = {'label': value[0], 'value': value[1]}
        return value

    def _handle_msg(self, msg):
        index = msg['item']
        if isinstance(index, list):
            index = tuple(index)
        value = self._lookup_item(index)
        if msg['type'] == 'click':
            if not isinstance(value, dict) or value.get('selectable', True):
                self.param.update(active=index, value=value)
            for fn in self._on_click_callbacks:
                try:
                    state.execute(partial(fn, value))
                except Exception as e:
                    print(f'List on_click handler errored: {e}')  # noqa
        elif msg['type'] == 'action':
            for fn in self._on_action_callbacks.get(msg['action'], []):
                try:
                    state.execute(partial(fn, value))
                except Exception as e:
                    print(f'List on_action handler errored: {e}')  # noqa

    def on_click(self, callback: Callable[[DOMEvent], None]):
        """
        Register a callback to be executed when a list item
        is clicked.

        Parameters
        ----------
        callback: (callable)
            The callback to run on click events.
        """
        self._on_click_callbacks.append(callback)

    def remove_on_click(self, callback: Callable[[DOMEvent], None]):
        """
        Remove a previously added click handler.

        Parameters
        ----------
        callback: (callable)
            The callback to run on edit events.
        """
        self._on_click_callbacks.remove(callback)


class Breadcrumbs(MenuBase):
    """
    The `Breadcrumbs` component is used to show the navigation path of a user within an application.
    It improves usability by allowing users to track their location and navigate back easily.

    Breadcrumb items can be strings or objects with properties:

    - `label`: The label of the breadcrumb item (required)
    - `icon`: The icon of the breadcrumb item (optional)
    - `avatar`: The avatar of the breadcrumb item (optional)
    - `href`: Link to navigate to when clicking the breadcrumb item (optional)

    :References:

    - https://panel-material-ui.holoviz.org/reference/menus/Breadcrumbs.html
    - https://mui.com/material-ui/react-breadcrumbs/

    :Example:

    >>> pmui.Breadcrumbs(items=[
    ...     {'label': 'Documentation', 'icon': 'article'},
    ...     {'label': 'Reference Gallery', 'icon': 'category'},
    ...     {'label': 'Menus', 'icon': 'menu'},
    ...     {'label': 'Breadcrumbs', 'icon': 'grain'},
    ... ], active=3)
    """

    color = param.Selector(objects=COLORS, default="primary")

    max_items = param.Integer(default=None, bounds=(1, None), doc="""
        The maximum number of breadcrumb items to display.""")

    separator = param.String(default=None, doc="""
        The separator displayed between breadcrumb items.""")

    _esm_base = "Breadcrumbs.jsx"
    _item_keys = ['label', 'icon', 'avatar', 'href']


class List(MenuBase):
    """
    The `List` component is used to display a structured group of items, such as menus,
    navigation links, or settings.

    List items can be strings or objects with properties:
      - `label`: The label of the list item (required)
      - `secondary`: The secondary text of the list item (optional)
      - `icon`: The icon of the list item (optional)
      - `avatar`: The avatar of the list item (optional)
      - `color`: The color of the list item (optional)
      - `actions`: Actions to display on the list item (optional)
      - `items`: Nested items (optional)
      - `selectable`: Whether the list item is selectable (optional)
      - `href`: The URL to navigate to when the list item is clicked (optional)

    :References:

    - https://panel-material-ui.holoviz.org/reference/menus/List.html
    - https://mui.com/material-ui/react-list/

    :Example:

    >>> pmui.List(items=[
    ...     {'label': 'Home', 'icon': 'home', 'secondary': 'Overview page'},
    ...     {'label': 'Gallery', 'icon': 'image', 'secondary': 'Visual overview'},
    ...     {'label': 'API', 'icon': 'code', 'secondary': 'API Reference'},
    ...     {'label': 'About', 'icon': 'info'},
    ... ], active=3)
    """

    active = param.ClassSelector(default=None, class_=(int, tuple), doc="""
        The index of the currently selected item. Can be a tuple of indices for nested items.""")

    color = param.Selector(default="default", objects=COLORS, doc="The color of the selected list item.")

    dense = param.Boolean(default=False, doc="Whether to show the list items in a dense format.")

    highlight = param.Boolean(default=True, doc="""
        Whether to highlight the currently selected menu item.""")

    level_indent = param.Integer(default=16, doc="The number of pixels to indent the list items.")

    removable = param.Boolean(default=False, doc="Whether to allow deleting items.")

    _esm_base = "List.jsx"

    _item_keys = ['label', 'items', 'icon', 'avatar', 'color', 'secondary', 'actions', 'selectable', 'href']

    @param.depends('items', watch=True, on_init=True)
    def _sync_items(self):
         pass

    def _process_property_change(self, props):
        props = super()._process_property_change(props)
        if 'active' in props and isinstance(props['active'], list):
            props['active'] = tuple(props['active'])
        return props

    def on_action(self, action: str, callback: Callable[[DOMEvent], None]):
        """
        Register a callback to be executed when an action is clicked.

        Parameters
        ----------
        action: (str)
            The action to register a callback for.
        callback: (callable)
            The callback to run on action events.
        """
        self._on_action_callbacks[action].append(callback)

    def remove_on_action(self, action: str, callback: Callable[[DOMEvent], None]):
        """
        Remove a previously added action handler.

        Parameters
        ----------
        action: (str)
            The action to remove a callback for.
        callback: (callable)
            The callback to remove.
        """
        self._on_action_callbacks[action].remove(callback)


class MenuButton(MenuBase, _ButtonBase):
    """
    The `MenuButton` component is a button component that allows selecting from a list of items.

    MenuButton items can be strings or objects with properties:
      - label: The label of the menu button item (required)
      - icon: The icon of the menu button item (optional)
      - color: The color of the menu button item (optional)

    :References:

    - https://panel-material-ui.holoviz.org/reference/menus/MenuButton.html
    - https://mui.com/material-ui/react-menu-button/

    :Example:

    >>> pmui.MenuButton(items=[
    ...     {'label': 'Open', 'icon': 'description'},
    ...     {'label': 'Save', 'icon': 'save'},
    ...     {'label': 'Exit', 'icon': 'close'},
    ... ], label='File', icon='storage')
    """

    margin = Margin(default=5)

    _esm_base = "MenuButton.jsx"
    _source_transforms = {
        "attached": None,
        "button_type": None,
        "button_style": None
    }
    _item_keys = ['label', 'icon', 'color', 'items']


class Pagination(MaterialWidget):
    """
    The `Pagination` component allows selecting from a list of pages.

    :References:

    - https://panel-material-ui.holoviz.org/reference/menus/Pagination.html
    - https://mui.com/material-ui/react-pagination/

    :Example:

    >>> pmui.Pagination(count=100)
    """

    boundary_count = param.Integer(default=1, bounds=(0, None), doc="The number of boundary pages to show.")

    color = param.Selector(default="default", objects=COLORS, doc="The color of the pagination.")

    count = param.Integer(default=1, bounds=(0, None), doc="The total number of pages.")

    shape = param.Selector(default="circular", objects=["circular", "rounded"], doc="The shape of the pagination.")

    size = param.Selector(default="medium", objects=["small", "medium", "large"], doc="The size of the pagination.")

    sibling_count = param.Integer(default=1, bounds=(0, None), doc="The number of sibling pages to show.")

    show_first_button = param.Boolean(default=False, doc="Whether to show the first button.")

    show_last_button = param.Boolean(default=False, doc="Whether to show the last button.")

    value = param.Integer(default=None, doc="The current zero-indexed page number.")

    variant = param.Selector(default="text", objects=["outlined", "text"], doc="The variant of the pagination.")

    width = param.Integer(default=None, doc="The width of the pagination.")

    _esm_base = "Pagination.jsx"

    @param.depends('count', watch=True, on_init=True)
    def _update_count(self):
        self.param.value.bounds = (0, self.count - 1 if self.count else 0)
        if self.count != 0 and self.value is None:
            self.value = 0

    @classmethod
    def paginate(cls, objects: list, layout: type[ListLike] = Column, page_size: int = 10, **params):
        """
        Paginate the items based on the current page and page size.

        Parameters
        ----------
        objects: list
            The list of objects to paginate.
        layout: type[LayoutBase]
            The layout to use for the paginated items.
        page_size: int
            The number of items to display per page.
        params: dict
            Additional parameters to pass to the layout.

        Returns
        -------
        layout
            The layout with the paginated items.
        """
        pagination = Pagination(count=len(objects)//page_size)
        val_rx = pagination.rx()
        objects_rx = param.rx(objects)[val_rx * page_size:(val_rx + 1) * page_size]
        return Column(
            layout(objects=objects_rx, **params),
            pagination
        )


class SpeedDial(MenuBase):
    """
    The `SpeedDial` component is a menu component that allows selecting from a
    list of items.

    SpeedDial items can be strings or objects with properties:

    - `label`: The label of the speed dial item (required)
    - `icon`: The icon of the speed dial item (optional)
    - `avatar`: The avatar of the speed dial item (optional)
    - `color`: The color of the speed dial item (optional)

    :References:

    - https://panel-material-ui.holoviz.org/reference/menus/SpeedDial.html
    - https://mui.com/material-ui/react-speed-dial/

    :Example:

    >>> pmui.SpeedDial(items=[
    ...     {'label': 'Camera', 'icon': 'camera'},
    ...     {'label': 'Photos', 'icon': 'photo'},
    ...     {'label': 'Documents', 'icon': 'article'},
    ... ], active=2, margin=(50, 20))
    """

    color = param.Selector(default="default", objects=COLORS, doc="""
        The color of the menu.""")

    direction = param.Selector(default="right", objects=["right", "left", "up", "down"], doc="""
        The direction of the menu.""")

    icon = param.String(default=None, doc="""
        The icon to display when the menu is closed.""")

    open_icon = param.String(default=None, doc="""
        The icon to display when the menu is open.""")

    persistent_tooltips = param.Boolean(default=False, doc="""
        Whether to show persistent tooltips next to the menu items.""")

    _esm_base = "SpeedDial.jsx"

    _item_keys = ['label', 'icon', 'avatar', 'color']



__all__ = [
    "Breadcrumbs",
    "List",
    "MenuButton",
    "Pagination",
    "SpeedDial"
]



import param
from panel.param import Param
from panel.widgets import WidgetBase

from .widgets import (
    Button,
    Checkbox,
    ColorPicker,
    DatePicker,
    DatetimeInput,
    DictInput,
    FileInput,
    FloatInput,
    FloatSlider,
    IntInput,
    IntSlider,
    ListInput,
    LiteralInput,
    MultiSelect,
    RangeSlider,
    Select,
    TextInput,
    TupleInput,
)


def SingleFileSelector(pobj: param.Parameter) -> type[WidgetBase]:
    """
    Determines whether to use a TextInput or Select widget for FileSelector
    """
    if pobj.path:
        return Select
    else:
        return TextInput


def LiteralInputTyped(pobj: param.Parameter) -> type[WidgetBase]:
    if isinstance(pobj, (param.Tuple, param.Range)):
        return TupleInput
    elif isinstance(pobj, param.Number):
        return type('NumberInput', (LiteralInput,), {'type': (int, float)})
    elif isinstance(pobj, param.Dict):
        return DictInput
    elif isinstance(pobj, param.List):
        return ListInput
    return LiteralInput


Param.mapping.update({
    param.Action:            Button,
    param.Boolean:           Checkbox,
    param.Bytes:             FileInput,
    param.CalendarDate:      DatePicker,
    param.Color:             ColorPicker,
    param.Date:              DatetimeInput,
    param.Dict:              LiteralInputTyped,
    param.FileSelector:      SingleFileSelector,
    param.Filename:          TextInput,
    param.Foldername:        TextInput,
    param.Integer:           IntSlider,
    param.List:              LiteralInputTyped,
    param.ListSelector:      MultiSelect,
    param.Number:            FloatSlider,
    param.ObjectSelector:    Select,
    param.Parameter:         LiteralInputTyped,
    param.Range:             RangeSlider,
    param.Selector:          Select,
    param.String:            TextInput,
})

Param.input_widgets.update({
    float: FloatInput,
    int: IntInput,
    "literal": LiteralInputTyped,
})



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

import param
from panel.chat.interface import CallbackState, ChatInterface
from panel.layout import Row

from .feed import ChatFeed
from .input import ChatAreaInput

if TYPE_CHECKING:
    pass

ICON_MAP = {
    "arrow-back": "undo",
    "trash": "delete",
}


class ChatInterface(ChatFeed, ChatInterface):
    """
    A chat interface that uses Material UI components.

    :References:

    - https://panel-material-ui.holoviz.org/reference/chat/ChatInterface.html
    - https://panel.holoviz.org/reference/chat/ChatInterface.html

    :Example:

    >>> ChatInterface().servable()
    """

    _input_type = ChatAreaInput

    _rename = {"loading": "loading"}

    @param.depends("_callback_state", watch=True)
    async def _update_input_disabled(self):
        busy_states = (CallbackState.RUNNING, CallbackState.GENERATING)
        if not self.show_stop or self._callback_state not in busy_states or self._callback_future is None:
            self._widget.loading = False
        else:
            self._widget.loading = True

    @param.depends("widgets", "button_properties", watch=True)
    def _init_widgets(self):
        if len(self.widgets) > 1:
            raise ValueError("panel_material_ui.ChatInterface.widgets not supported.")
        self._init_button_data()
        actions = {}
        for name, data in self._button_data.items():
            if (
                name in ("send", "stop") or (name == "rerun" and not self.show_rerun) or
                (name == "undo" and not self.show_undo) or (name == "clear" and not self.show_clear)
            ):
                continue
            actions[name] = {'icon': ICON_MAP.get(data.icon, data.icon), 'callback': partial(data.callback, self), 'label': name.title()}
        self._widget = ChatAreaInput(actions=actions, sizing_mode="stretch_width")
        self.link(self._widget, disabled="disabled_enter")
        callback = partial(self._button_data["send"].callback, self)
        self._widget.param.watch(callback, "value")
        self._widget.on_action("stop", self._click_stop)
        input_container = Row(self._widget, sizing_mode="stretch_width")
        self._input_container.objects = [input_container]
        self._input_layout = input_container

    @param.depends("placeholder_text", "placeholder_params", watch=True, on_init=True)
    def _update_placeholder(self):
        self._placeholder = self._message_type(
            self.placeholder_text,
            avatar='PLACEHOLDER',
            css_classes=["message"],
            **self.placeholder_params
        )

__all__ = ["ChatInterface"]



from __future__ import annotations

import colorsys

import numpy as np
import param
from bokeh.themes import Theme as _BkTheme
from panel.config import config
from panel.theme.material import Material, MaterialDarkTheme, MaterialDefaultTheme
from panel.viewable import Viewable

MATERIAL_UI_ICONS = """
.material-icons { font-family: 'Material Icons'; }
.material-icons-outlined { font-family: 'Material Icons Outlined'; }
"""

MATERIAL_THEME = {
    "attrs": {
        "Axis": {
            "axis_label_standoff": 10,
            "axis_label_text_font_size": "1.25em",
            "axis_label_text_font_style": "normal",
            "major_label_text_font_size": "1.025em",
        },
        "Legend": {
            "spacing": 8,
            "glyph_width": 15,
            "label_standoff": 8,
            "label_text_font_size": "1.025em",
        },
        "ColorBar": {
            "title_text_font_size": "1.025em",
            "title_text_font_style": "normal",
            "major_label_text_font_size": "1.025em",
        },
        "Title": {
            "text_font_size": "1.15em",
        },
    }
}

def rgb2hex(rgb: tuple[int, int, int]) -> str:
    """
    Convert RGB tuple to hex.

    Parameters
    ----------
    rgb : tuple
        The RGB(A) tuple to convert.

    Returns
    -------
    str
        The hex color.
    """
    return "#{:02x}{:02x}{:02x}".format(*(int(v*255) for v in rgb))

def hex2rgb(hex: str) -> tuple[int, int, int]:
    """
    Convert hex color to RGB tuple.

    Parameters
    ----------
    hex : str
        The hex color to convert.

    Returns
    -------
    tuple[int, int, int]
        The RGB tuple.
    """
    return [int(hex[i : i + 2], 16) for i in range(1, 6, 2)]

def linear_gradient(start_hex: str, finish_hex: str, n: int = 10) -> list[str]:
    """
    Interpolates the color gradient between to hex colors.

    Parameters
    ----------
    start_hex : str
        The starting hex color.
    finish_hex : str
        The finishing hex color.
    n : int, optional
        The number of colors to generate.

    Returns
    -------
    list[str]
        A list of colors in hex format.
    """
    s = hex2rgb(start_hex)
    f = hex2rgb(finish_hex)
    gradient = [s]
    for t in range(1, n):
        curr_vector = [int(s[j] + (float(t)/(n-1))*(f[j]-s[j])) for j in range(3)]
        gradient.append(curr_vector)
    return [rgb2hex([c/255. for c in rgb]) for rgb in gradient]

def generate_palette(color: str, n_colors: int = 3) -> list[str]:
    """
    Generate a palette of colors from a base color.

    Parameters
    ----------
    color : str
        The base color to generate the palette from.
    n_colors : int, optional
        The number of colors to generate.

    Returns
    -------
    list[str]
        A list of colors in hex format.
    """
    hex_color = color.lstrip("#")
    r, g, b = tuple(int(hex_color[i : i + 2], 16) / 255.0 for i in (0, 2, 4))
    h, l, s = colorsys.rgb_to_hls(r, g, b)  # noqa

    hues = np.linspace(0, 1, int(n_colors) + 1)[:-1]
    hues += h
    hues %= 1
    hues -= hues.astype(int)
    rgb_palette = [colorsys.hls_to_rgb(h_i, l, s) for h_i in hues]
    hex_palette = [
        f"#{int(r_i * 255):02x}{int(g_i * 255):02x}{int(b_i * 255):02x}"
        for r_i, g_i, b_i in rgb_palette
    ]
    return hex_palette


class MuiDefaultTheme(MaterialDefaultTheme):

    bokeh_theme = param.ClassSelector(
        class_=(_BkTheme, str), default=_BkTheme(json=MATERIAL_THEME))

class MuiDarkTheme(MaterialDarkTheme):

    bokeh_theme = param.ClassSelector(
        class_=(_BkTheme, str), default=_BkTheme(json=MATERIAL_THEME))


class MaterialDesign(Material):

    modifiers = {
        **Material.modifiers,
        Viewable: {
            'stylesheets': Material.modifiers[Viewable]['stylesheets'] + [MATERIAL_UI_ICONS]
        }
    }

    _resources = {}
    _themes = {'dark': MuiDarkTheme, 'default': MuiDefaultTheme}

    @classmethod
    def _get_modifiers(cls, viewable, theme, isolated=False):
        modifiers, child_modifiers = super()._get_modifiers(viewable, theme, isolated)
        if hasattr(viewable, '_esm_base'):
            del modifiers['stylesheets']
        return modifiers, child_modifiers


param.Parameterized.__setattr__(config, 'design', MaterialDesign)



import numpy as np
import pytest

from pathlib import Path
from panel import config
from datetime import date, datetime, time as dt_time

from panel_material_ui.widgets import FileInput, IntInput, FloatInput, DatePicker



@pytest.mark.from_panel
def test_file_input(document, comm):
    file_input = FileInput(accept='.txt')

    file_input._process_events({'mime_type': 'text/plain', 'value': 'U29tZSB0ZXh0Cg==', 'filename': 'testfile'})
    assert file_input.value == b'Some text\n'
    assert file_input.mime_type == 'text/plain'
    assert file_input.accept == '.txt'
    assert file_input.filename == 'testfile'


@pytest.mark.from_panel
def test_file_input_save_one_file(document, comm, tmpdir):
    file_input = FileInput(accept='.txt')

    file_input._process_events({'mime_type': 'text/plain', 'value': 'U29tZSB0ZXh0Cg==', 'filename': 'testfile'})

    fpath = Path(tmpdir) / 'out.txt'
    file_input.save(str(fpath))

    assert fpath.exists()
    content = fpath.read_text()
    assert content == 'Some text\n'


@pytest.mark.from_panel
@pytest.mark.xfail(reason='')
def test_int_input(document, comm):
    int_input = IntInput(name='Int input')
    widget = int_input.get_root(document, comm=comm)

    assert widget.name == 'Int input'
    assert widget.step == 1
    assert widget.value == 0

    int_input._process_events({'value': 2})
    assert int_input.value == 2
    int_input._process_events({'value_throttled': 2})
    assert int_input.value_throttled == 2

    int_input.value = 0
    assert widget.value == 0

    # Testing throttled mode
    with config.set(throttled=True):
        int_input._process_events({'value': 1})
        assert int_input.value == 0  # no change
        int_input._process_events({'value_throttled': 1})
        assert int_input.value == 1

        int_input.value = 2
        assert widget.value == 2


@pytest.mark.from_panel
@pytest.mark.xfail(reason='')
def test_float_input(document, comm):
    float_input = FloatInput(value=0.4, name="Float input")
    widget = float_input.get_root(document, comm=comm)

    assert widget.name == 'Float input'
    assert widget.step == 0.1
    assert widget.value == 0.4

    float_input._process_events({'value': 0.2})
    assert float_input.value == 0.2
    float_input._process_events({'value_throttled': 0.2})
    assert float_input.value_throttled == 0.2

    float_input.value = 0.3
    assert widget.value == 0.3

    # Testing throttled mode
    with config.set(throttled=True):
        float_input._process_events({'value': 0.4})
        assert float_input.value == 0.3  # no change
        float_input._process_events({'value_throttled': 0.4})
        assert float_input.value == 0.4

        float_input.value = 0.5
        assert widget.value == 0.5


@pytest.mark.from_panel
def test_date_picker():
    date_picker = DatePicker(name='DatePicker', value=date(2018, 9, 2),
                             start=date(2018, 9, 1), end=date(2018, 9, 10))

    date_picker._process_events({'value': '2018-09-03'})
    assert date_picker.value == date(2018, 9, 3)

    date_picker._process_events({'value': date(2018, 9, 5)})
    assert date_picker.value == date(2018, 9, 5)

    date_picker._process_events({'value': date(2018, 9, 6)})
    assert date_picker.value == date(2018, 9, 6)


@pytest.mark.from_panel
def test_date_picker_options():
    options = [date(2018, 9, 1), date(2018, 9, 2), date(2018, 9, 3)]
    datetime_picker = DatePicker(
        name='DatetimePicker', value=date(2018, 9, 2),
        options=options
    )
    assert datetime_picker.value == date(2018, 9, 2)
    assert datetime_picker.enabled_dates == options



from __future__ import annotations

import datetime as dt

import param
from bokeh.models.formatters import NumeralTickFormatter, TickFormatter
from panel.util import datetime_as_utctimestamp, edit_readonly, value_as_date, value_as_datetime
from panel.widgets.select import SingleSelectBase as _PnSingleSelectBase
from panel.widgets.slider import _SliderBase
from param.parameterized import resolve_value

from ..base import COLORS
from .base import MaterialWidget


class _ContinuousSlider(MaterialWidget, _SliderBase):

    bar_color = param.Color(default=None, allow_None=True, doc="Color of the bar")

    color = param.Selector(objects=COLORS, default="default")

    direction = param.Selector(default='ltr', objects=['ltr', 'rtl'], doc="""
        Whether the slider should go from left-to-right ('ltr') or
        right-to-left ('rtl').""")

    start = param.Number(default=0)

    end = param.Number(default=100)

    format = param.ClassSelector(default='0[.]00', class_=(str, TickFormatter,), doc="""
        A custom format string or Bokeh TickFormatter.""")

    marks = param.ClassSelector(class_=(bool, list), default=False, doc="""
        Marks indicate predetermined values to which the user can move the slider.
        If True the `options` are shown as marks. If a list, it should contain dicts with 'value'
        and an optional 'label' keys.""")

    size = param.Selector(objects=["small", "medium", "large"], default="medium")

    step = param.Number(default=1)

    tooltips = param.Selector(objects=[True, False, "auto"], default="auto", doc="""
        Whether the slider handle should display tooltips (if auto will render on hover).""")

    track = param.Selector(objects=["normal", "inverted", False], default="normal")

    value = param.Number(default=0)

    value_label = param.String(default=None)

    value_throttled = param.Number(default=0, constant=True)

    _esm_base = "Slider.jsx"

    _rename = {"name": "name"}

    __abstract = True

    def _process_param_change(self, params):
        if self.orientation == 'vertical' and ('width' in params or 'height' in params):
            params['width'] = self.height
            params['height'] = self.width
        return super()._process_param_change(params)


class IntSlider(_ContinuousSlider):
    """
    The IntSlider widget allows selecting an integer value within a
    set of bounds using a slider.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/IntSlider.html
    - https://panel.holoviz.org/reference/widgets/IntSlider.html
    - https://mui.com/material-ui/react-slider/

    :Example:

    >>> IntSlider(value=5, start=0, end=10, step=1, label="Integer Value")
    """

    end = param.Integer(default=1)

    format = param.ClassSelector(default='0,0', class_=(str, TickFormatter,), doc="""
        A custom format string or Bokeh TickFormatter.""")

    start = param.Integer(default=1)

    step = param.Integer(default=1, bounds=(1, None))

    value = param.Integer(default=0)

    value_throttled = param.Integer(default=0, constant=True)


class FloatSlider(_ContinuousSlider):
    """
    The FloatSlider widget allows selecting a floating-point value
    within a set of bounds using a slider.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/FloatSlider.html
    - https://panel.holoviz.org/reference/widgets/FloatSlider.html
    - https://mui.com/material-ui/react-slider/

    :Example:

    >>> FloatSlider(value=0.5, start=0.0, end=1.0, step=0.1, label="Float value")
    """

    step = param.Number(default=0.1, doc="The step size.")


class DateSlider(_ContinuousSlider):
    """
    The DateSlider widget allows selecting a value within a set of
    bounds using a slider.  Supports datetime.datetime, datetime.date
    and np.datetime64 values. The step size is fixed at 1 day.

    :References:

    - https://panel.holoviz.org/reference/widgets/DateSlider.html

    :Example:

    >>> import datetime as dt
    >>> DateSlider(
    ...     value=dt.datetime(2025, 1, 1),
    ...     start=dt.datetime(2025, 1, 1),
    ...     end=dt.datetime(2025, 1, 7),
    ...     name="A datetime value"
    ... )
    """

    as_datetime = param.Boolean(default=False, doc="""
        Whether to store the date as a datetime.""")

    end = param.Date(default=None, doc="""
        The upper bound.""")

    format = param.String(default=None, doc="""
        Datetime format used for parsing and formatting the date.""")

    start = param.Date(default=None, doc="""
        The lower bound.""")

    step = param.Integer(default=1, bounds=(1, None), doc="""
        The step parameter in days.""")

    value = param.Date(default=None, doc="""
        The selected date value of the slider. Updated when the slider
        handle is dragged. Supports datetime.datetime, datetime.date
        or np.datetime64 types.""")

    value_throttled = param.Date(default=None, constant=True, doc="""
        The value of the slider. Updated when the slider handle is released.""")

    _constants = {"date": True}
    _rename = {"as_datetime": None}
    _source_transforms = {
        "value": None, "value_throttled": None, "value_start": None,
        "value_end": None, "start": None, "end": None
    }

    def _process_param_change(self, msg):
        msg = super()._process_param_change(msg)
        if 'value' in msg:
            value = msg['value']
            if isinstance(value, dt.datetime):
                value = datetime_as_utctimestamp(value)
            msg['value'] = value
        return msg

    def _process_property_change(self, msg):
        msg = super()._process_property_change(msg)
        transform = value_as_datetime if self.as_datetime else value_as_date
        if 'value' in msg:
            msg['value'] = transform(msg['value'])
        if 'value_throttled' in msg:
            msg['value_throttled'] = transform(msg['value_throttled'])
        return msg


class DatetimeSlider(DateSlider):
    """
    The DatetimeSlider widget allows selecting a value within a set of
    bounds using a slider. Supports datetime.date, datetime.datetime
    and np.datetime64 values. The step size is fixed at 1 minute.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/DatetimeSlider.html
    - https://panel.holoviz.org/reference/widgets/DatetimeSlider.html
    - https://mui.com/material-ui/react-slider/

    :Example:

    >>> import datetime as dt
    >>> DatetimeSlider(
    ...     value=dt.datetime(2025, 1, 1),
    ...     start=dt.datetime(2025, 1, 1),
    ...     end=dt.datetime(2025, 1, 7),
    ...     name="A datetime value"
    ... )
    """

    as_datetime = param.Boolean(default=True, readonly=True, doc="""
        Whether to store the date as a datetime.""")

    step = param.Number(default=60, bounds=(1, None), doc="""
        The step size in seconds. Default is 1 minute, i.e 60 seconds.""")

    _property_conversion = staticmethod(value_as_datetime)

    _constants = {"datetime": True}


class _RangeSliderBase(_ContinuousSlider):

    value = param.Range(default=(0, 100))

    value_throttled = param.Range(default=(0, 100), readonly=True)

    value_start = param.Parameter(readonly=True, doc="""The lower value of the selected range.""")

    value_end = param.Parameter(readonly=True, doc="""The upper value of the selected range.""")

    __abstract = True

    def __init__(self, **params):
        if "value" not in params:
            params["value"] = (params.get("start", self.start), params.get("end", self.end))
        if params["value"] is not None:
            v1, v2 = params["value"]
            params["value_start"], params["value_end"] = resolve_value(v1), resolve_value(v2)
        if 'format' in params and isinstance(params['format'], str):
            params['format'] = NumeralTickFormatter(format=params['format'])
        with edit_readonly(self):
            super().__init__(**params)

    def _process_property_change(self, msg):
        msg = super()._process_property_change(msg)
        if 'value' in msg:
            msg['value'] = tuple(msg['value'])
        if 'value_throttled' in msg:
            msg['value_throttled'] = tuple(msg['value_throttled'])
        return msg

    @param.depends("value", watch=True)
    def _sync_values(self):
        vs, ve = self.value
        with edit_readonly(self):
            self.param.update(value_start=vs, value_end=ve)


class RangeSlider(_RangeSliderBase):
    """
    The RangeSlider widget allows selecting a floating-point range
    using a slider with two handles.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/RangeSlider.html
    - https://panel.holoviz.org/reference/widgets/RangeSlider.html
    - https://mui.com/material-ui/react-slider/

    :Example:

    >>> RangeSlider(
    ...     value=(1.0, 1.5), start=0.0, end=2.0, step=0.25, label="A tuple of floats"
    ... )
    """


class IntRangeSlider(_RangeSliderBase):
    """
    The IntRangeSlider widget allows selecting an integer range using
    a slider with two handles.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/IntRangeSlider.html
    - https://panel.holoviz.org/reference/widgets/IntRangeSlider.html
    - https://mui.com/material-ui/react-slider/

    :Example:

    >>> IntRangeSlider(
    ...     value=(2, 4), start=0, end=10, step=2, label="A tuple of integers"
    ... )
    """

    start = param.Integer(default=0)

    format = param.ClassSelector(default='0,0', class_=(str, TickFormatter,), doc="""
        A custom format string or Bokeh TickFormatter.""")

    end = param.Integer(default=100)

    step = param.Integer(default=1)

    value_start = param.Integer(default=0, readonly=True, doc="""The lower value of the selected range.""")

    value_end = param.Integer(default=100, readonly=True, doc="""The upper value of the selected range.""")


class DateRangeSlider(_RangeSliderBase):
    """
    The DateRangeSlider widget allows selecting a date range using a
    slider with two handles. Supports datetime.datetime, datetime.date
    and np.datetime64 ranges.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/DateRangeSlider.html
    - https://panel.holoviz.org/reference/widgets/DateRangeSlider.html
    - https://mui.com/material-ui/react-slider/

    :Example:

    >>> import datetime as dt
    >>> DateRangeSlider(
    ...     value=(dt.datetime(2025, 1, 9), dt.datetime(2025, 1, 16)),
    ...     start=dt.datetime(2025, 1, 1),
    ...     end=dt.datetime(2025, 1, 31),
    ...     step=2,
    ...     name="A tuple of datetimes"
    ... )
    """

    value = param.DateRange(default=None, allow_None=False, doc="""
        The selected range as a tuple of values. Updated when one of the handles is
        dragged. Supports datetime.datetime, datetime.date, and np.datetime64 ranges.""")

    value_start = param.Date(default=None, readonly=True, doc="""
        The lower value of the selected range.""")

    value_end = param.Date(default=None, readonly=True, doc="""
        The upper value of the selected range.""")

    value_throttled = param.DateRange(default=None, constant=True, nested_refs=True, doc="""
        The selected range as a tuple of values. Updated one of the handles is released. Supports
        datetime.datetime, datetime.date and np.datetime64 ranges""")

    start = param.Date(default=None, doc="""
        The lower bound.""")

    end = param.Date(default=None, doc="""
        The upper bound.""")

    step = param.Number(default=1, doc="""
        The step size in days. Default is 1 day.""")

    format = param.String(default=None, doc="""
        Datetime format used for parsing and formatting the date.""")

    _constants = {"date": True}

    _property_conversion = staticmethod(value_as_date)

    _rename = {'value_start': None, 'value_end': None}
    _source_transforms = {
        "value": None, "value_throttled": None, "value_start": None,
        "value_end": None, "start": None, "end": None
    }

    def _process_param_change(self, msg):
        msg = super()._process_param_change(msg)
        if msg.get('value', 'unchanged') is None:
            del msg['value']
        elif 'value' in msg:
            v1, v2 = msg['value']
            if isinstance(v1, dt.datetime):
                v1 = datetime_as_utctimestamp(v1)
            if isinstance(v2, dt.datetime):
                v2 = datetime_as_utctimestamp(v2)
            msg['value'] = (v1, v2)
        if msg.get('value_throttled', 'unchanged') is None:
            del msg['value_throttled']
        return msg

    def _process_property_change(self, msg):
        msg = super()._process_property_change(msg)
        if 'value' in msg:
            v1, v2 = msg['value']
            msg['value'] = (self._property_conversion(v1), self._property_conversion(v2))
        if 'value_throttled' in msg:
            v1, v2 = msg['value_throttled']
            msg['value_throttled'] = (self._property_conversion(v1), self._property_conversion(v2))
        return msg


class DatetimeRangeSlider(DateRangeSlider):

    """
    The DatetimeRangeSlider widget allows selecting a datetime range
    using a slider with two handles. Supports datetime.datetime and
    np.datetime64 ranges.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/DatetimeRangeSlider.html
    - https://panel.holoviz.org/reference/widgets/DatetimeRangeSlider.html
    - https://mui.com/material-ui/react-slider/

    :Example:

    >>> import datetime as dt
    >>> DatetimeRangeSlider(
    ...     value=(dt.datetime(2025, 1, 9), dt.datetime(2025, 1, 16)),
    ...     start=dt.datetime(2025, 1, 1),
    ...     end=dt.datetime(2025, 1, 31),
    ...     step=60*60,
    ...     label="A tuple of datetimes"
    ... )
    """

    step = param.Number(default=60, doc="""
        The step size in seconds. Default is 1 minute, i.e 60 seconds.""")

    _property_conversion = staticmethod(value_as_datetime)

    _constants = {"datetime": True}


class DiscreteSlider(IntSlider, _PnSingleSelectBase):
    """
    The DiscreteSlider widget allows selecting a discrete value using a slider.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/DiscreteSlider.html
    - https://panel.holoviz.org/reference/widgets/DiscreteSlider.html
    - https://mui.com/material-ui/react-slider/
    """

    options = param.ClassSelector(default=[], class_=(dict, list), doc="""
        A list or dictionary of valid options.""")

    value = param.Parameter(doc="""
        The selected value of the slider. Updated when the handle is
        dragged. Must be one of the options.""")

    value_throttled = param.Parameter(constant=True, doc="""
        The value of the slider. Updated when the handle is released.""")

    start = param.Integer(default=0, readonly=True)
    end = param.Integer(default=100, readonly=True)
    step = param.Integer(default=1, readonly=True)

    _allows_values = False
    _constants = {"discrete": True}

    @param.depends("options", watch=True)
    def _update_bounds(self):
        self.param.update(start=0, end=len(self.options)-1)

    def _process_param_change(self, msg):
        msg = super()._process_param_change(msg)
        if 'options' in msg:
            msg['options'] = self.labels
        if 'value' in msg:
            msg['value'] = self.labels.index(msg['value'])
        return msg

    def _process_property_change(self, msg):
        if 'value' in msg:
            msg['value'] = self.values[msg['value']]
        msg = super()._process_property_change(msg)
        return msg


class Rating(MaterialWidget):
    """
    The Rating slider widget allows users to select a rating value of their own.

    :References:

    - https://mui.com/material-ui/react-rating/

    :Example:

    >>> Rating(value=3, size="large", name="Rate the product")
    """

    end = param.Integer(default=5, bounds=(1, None), doc="The maximum value for the rating.")

    only_selected = param.Boolean(default=False, doc="Whether to highlight only the select value")

    size = param.Selector(default="medium", objects=["small", "medium", "large"])

    value = param.Number(default=0, allow_None=True, bounds=(0, 5))

    _esm_base = "Rating.jsx"

    @param.depends("end", watch=True, on_init=True)
    def _update_value_bounds(self):
        self.param.value.bounds = (0, self.end)

    def _process_property_change(self, msg):
        if 'value' in msg and msg['value'] is None:
            msg['value'] = 0
        return super()._process_property_change(msg)


class _EditableContinuousSliderBase(_ContinuousSlider):

    _constants = {"editable": True}
    _esm_base = "Slider.jsx"


class EditableFloatSlider(_EditableContinuousSliderBase, FloatSlider):
    """
    The EditableFloatSlider widget allows selecting a
    numeric floating-point value within a set of bounds using a slider
    and for more precise control offers an editable number input box.

    :References:
    - https://panel-material-ui.holoviz.org/reference/widgets/EditableFloatSlider.html
    - https://panel.holoviz.org/reference/widgets/EditableFloatSlider.html
    - https://mui.com/material-ui/react-slider/

    :Example:

    >>> EditableFloatSlider(
    ...     value=1.0, start=0.0, end=2.0, step=0.25, label="A float value"
    ... )
    """

    fixed_start = param.Number(default=None, doc="""
        A fixed lower bound for the slider and input.""")

    fixed_end = param.Number(default=None, doc="""
        A fixed upper bound for the slider and input.""")


class EditableIntSlider(_EditableContinuousSliderBase, IntSlider):
    """
    The EditableIntSlider widget allows selecting an integer
    value within a set of bounds using a slider and for more precise
    control offers an editable integer input box.

    :References:

    - https://panel-material-ui.holoviz.org/reference/widgets/EditableIntSlider.html
    - https://panel.holoviz.org/reference/widgets/EditableIntSlider.html
    - https://mui.com/material-ui/react-slider/

    :Example:

    >>> EditableIntSlider(
    ...     value=2, start=0, end=5, step=1, label="An integer value"
    ... )
    """

    fixed_start = param.Integer(default=None, doc="""
        A fixed lower bound for the slider and input.""")

    fixed_end = param.Integer(default=None, doc="""
       A fixed upper bound for the slider and input.""")


__all__ = [
    "DateSlider",
    "DatetimeSlider",
    "DateRangeSlider",
    "DatetimeRangeSlider",
    "DiscreteSlider",
    "EditableFloatSlider",
    "EditableIntSlider",
    "FloatSlider",
    "IntRangeSlider",
    "IntSlider",
    "RangeSlider",
    "Rating"
]



import numpy as np
import pytest
from panel.pane import panel
from panel_material_ui.widgets import AutocompleteInput, Select


@pytest.mark.parametrize('widget', [AutocompleteInput, Select])
def test_list_constructor(widget):
    select = widget(options=['A', 1], value=1)
    assert select.options == ['A', 1]

@pytest.mark.parametrize('widget', [AutocompleteInput, Select])
def test_select_float_option_with_equality(widget):
    opts = {'A': 3.14, '1': 2.0}
    select = widget(options=opts, value=3.14, name='Select')
    assert select.value == 3.14

    select.value = 2
    assert select.value == 2.0

    select.value = 3.14
    assert select.value == 3.14

@pytest.mark.parametrize('widget', [AutocompleteInput, Select])
def test_select_text_option_with_equality(widget):
    opts = {'A': 'ABC', '1': 'DEF'}
    select = widget(options=opts, value='DEF', name='Select')
    assert select.value == 'DEF'

    select.value = 'ABC'
    assert select.value == 'ABC'

    select.value = 'DEF'
    assert select.value == 'DEF'

def test_autocomplete(document, comm):
    opts = {'A': 'a', '1': 1}
    select = AutocompleteInput(options=opts, value=opts['1'], name='Autocomplete')

    widget = select.get_root(document, comm=comm)

    assert widget.data.label == 'Autocomplete'
    assert widget.data.value == str(opts['1'])
    assert widget.data.options == list(opts)

    select._process_events({'value': 'A'})
    assert select.value == 'a'

    widget.data.value = '1'
    select.value = opts['1']
    assert select.value == opts['1']

    select.value = opts['A']
    assert widget.data.value == 'A'

def test_autocomplete_reset_none(document, comm):
    widget = AutocompleteInput(options=['A', 'B', 'C'], value='B')

    model = widget.get_root(document, comm=comm)

    assert widget.value == model.data.value == 'B'

    model.data.value = None

    assert widget.value is None

def test_autocomplete_unrestricted(document, comm):
    opts = {'A': 'a', '1': 1}
    select = AutocompleteInput(options=opts, value=opts['1'], name='Autocomplete', restrict=False)

    widget = select.get_root(document, comm=comm)

    assert widget.data.label == 'Autocomplete'
    assert widget.data.value == str(opts['1'])
    assert widget.data.options == list(opts)

    select._process_events({'value': str(opts['A'])})
    assert select.value == opts['A']

    select._process_events({'value': 'foo'})
    assert select.value == 'foo'

    select.value = 'bar'
    assert widget.data.value == 'bar'

def test_select(document, comm):
    opts = {'A': 'a', '1': 1}
    select = Select(options=opts, value=opts['1'], name='Select')

    widget = select.get_root(document, comm=comm)

    assert widget.data.label == 'Select'
    assert widget.data.value == str(opts['1'])
    assert widget.data.options == list(opts)

    select._process_events({'value': str(opts['A'])})
    assert select.value == opts['A']

    widget.data.value = str(opts['1'])
    select.value = opts['1']
    assert select.value == opts['1']

    select.value = opts['A']
    assert widget.data.value == 'A'

def test_select_groups_list_options(document, comm):
    groups = dict(a=[1, 2], b=[3])
    select = Select(value=groups['a'][0], groups=groups, name='Select')

    widget = select.get_root(document, comm=comm)

    assert widget.data.label == 'Select'
    assert widget.data.value == str(groups['a'][0])
    assert widget.data.options == {gr: [(str(v), str(v)) for v in values] for gr, values in groups.items()}

    select._process_events({'value': str(groups['a'][1])})
    assert select.value == groups['a'][1]

    select._process_events({'value': str(groups['a'][0])})
    assert select.value == groups['a'][0]

    select.value = groups['a'][1]
    assert widget.data.value == str(groups['a'][1])

def test_select_groups_dict_options(document, comm):
    groups = dict(A=dict(a=1, b=2), B=dict(c=3))
    select = Select(value=groups['A']['a'], groups=groups, name='Select')

    widget = select.get_root(document, comm=comm)

    assert widget.data.label == 'Select'
    assert widget.data.value == 'a'
    assert widget.data.options == {'A': [('1', 'a'), ('2', 'b')], 'B': [('3', 'c')]}

    select._process_events({'value': str(groups['B']['c'])})
    assert select.value == groups['B']['c']

    select._process_events({'value': str(groups['A']['b'])})
    assert select.value == groups['A']['b']

    select.value = groups['A']['a']
    assert widget.data.value == 'a'

def test_select_change_groups(document, comm):
    groups = dict(A=dict(a=1, b=2), B=dict(c=3))
    select = Select(value=groups['A']['a'], groups=groups, name='Select')

    widget = select.get_root(document, comm=comm)

    new_groups = dict(C=dict(d=4), D=dict(e=5, f=6))
    select.groups = new_groups
    assert select.value == new_groups['C']['d']
    assert widget.data.value == 'd'
    assert widget.data.options == {'C': [('4', 'd')], 'D': [('5', 'e'), ('6', 'f')]}

    select.groups = {}
    assert select.value is None
    assert widget.data.value is None

def test_select_groups_error_with_options():
    # Instantiate with groups and options
    with pytest.raises(ValueError):
        Select(options=[1, 2], groups=dict(a=[1], b=[2]), name='Select')

    opts = [1, 2, 3]
    groups = dict(a=[1, 2], b=[3])

    # Instamtiate with options and then update groups
    select = Select(options=opts, name='Select')
    with pytest.raises(ValueError):
        select.groups = groups

    # Instantiate with groups and then update options
    select = Select(groups=groups, name='Select')
    with pytest.raises(ValueError):
        select.options = opts

@pytest.mark.parametrize('widget', [AutocompleteInput, Select])
def test_select_change_options(widget, document, comm):
    opts = {'A': 'a', '1': 1}
    select = widget(options=opts, value=opts['1'], name='Select')

    widget = select.get_root(document, comm=comm)

    select.options = {'A': 'a'}
    if select._allows_none:
        assert select.value is None
        assert widget.data.value is None
    else:
        assert select.value == opts['A']
        assert widget.data.value == 'A'

    select.options = {}
    assert select.value is select.param['value'].default
    assert widget.data.value == select.param['value'].default

@pytest.mark.parametrize('widget', [AutocompleteInput, Select])
def test_select_non_hashable_options(widget, document, comm):
    opts = {'A': np.array([1, 2, 3]), '1': np.array([3, 4, 5])}
    select = widget(options=opts, value=opts['1'], name='Select')

    widget = select.get_root(document, comm=comm)

    if select._allows_none:
        select.value = None
        assert select.value is None
        assert widget.data.value is None
    else:
        select.value = opts['A']
        assert select.value is opts['A']
        assert widget.data.value == (str(opts['A']) if select._allows_values else 'A')

    opts.pop('A')
    select.options = opts
    if select._allows_none:
        assert select.value is None
        assert widget.data.value is None
    else:
        assert select.value is opts['1']
        assert widget.data.value == '1'

def test_select_mutables(document, comm):
    opts = {'A': [1,2,3], 'B': [2,4,6], 'C': dict(a=1,b=2)}
    select = Select(options=opts, value=opts['B'], name='Select')

    widget = select.get_root(document, comm=comm)

    assert widget.data.label == 'Select'
    assert widget.data.value == 'B'
    assert widget.data.options == list(opts)

    widget.data.value = str(opts['B'])
    select._process_events({'value': str(opts['A'])})
    assert select.value == opts['A']

    widget.data.value = str(opts['B'])
    select._process_events({'value': str(opts['B'])})
    assert select.value == opts['B']

    select.value = opts['A']
    assert widget.data.value == 'A'

def test_select_change_options_on_watch(document, comm):
    select = Select(options={'A': 'A', '1': 1, 'C': object},
                    value='A', name='Select')

    def set_options(event):
        if event.new == 1:
            select.options = {'D': 2, 'E': 'a'}
    select.param.watch(set_options, 'value')

    model = select.get_root(document, comm=comm)

    select.value = 1
    assert select.value == 2
    assert model.data.value == 'D'
    assert model.data.options == list(select.options)
